/**
  *  \file client/cm-marker.cc
  *  \brief Chart Modes dealing with Drawings
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/cm-marker.h"
#include "client/chartloc.h"
#include "client/dlg-draw.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "ui/widgets.h"
#include "game/global.h"

/**************************** WMarkerChartMode ***************************/

// FIXME: observe current_turn, and automatically exit this mode
// when current_turn changes. Likewise, observe GDrawingContainer::sig_deleted
// and exit this mode when our drawing is deleted

/** Constructor. */
WMarkerChartMode::WMarkerChartMode(GDrawingContainer::iterator drawing)
    : WSimpleChartMode(ChartMode_PrimaryMode),
      drawing(drawing)
{ }

WMarkerChartMode::~WMarkerChartMode()
{ }

bool
WMarkerChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_ESCAPE:
         case vk_Quit:
            // Exit move mode
            delete this;
            return true;
         case 'c':
            // Change color
            {
                WColorSelector csel;
                csel.setColor(GDrawing::getUserColorFromColor(drawing->getColor()));
                if (csel.doStandardDialog(_("Drawing Color"), false) == csel.rOK) {
                    drawing->setColor(GDrawing::getColorFromUserColor(csel.getColor()));
                    loc.markChanged();
                }
            }
            return true;
         case SDLK_DELETE:
            // Delete marker
            {
                GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
                c.erase(drawing);
                drawing = c.end();
            }
            delete this;
            return true;
        }
    }
    return WSimpleChartMode::handleEvent(event, loc);
}

/** Start moving a marker. Creates a WMoveMarkerChartMode and adds it to \c modes. */
void
WMarkerChartMode::moveMarker(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WMoveMarkerChartMode(p));
}

/** Start drawing a line or rectangle. Creates a WDrawMarkerChartMode
    and adds it to \c modes. */
void
WMarkerChartMode::drawLineRectangle(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WDrawMarkerChartMode(p));
}

/** Start drawing a circle. Creates a WDrawCircleChartMode and adds it to \c modes. */
void
WMarkerChartMode::drawCircle(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WDrawCircleChartMode(p));
}

/** Start deleting a drawing. Creates a WDeleteDrawingChartMode and adds it to \c modes. */
void
WMarkerChartMode::deleteDrawing(WChartModeList& modes, GDrawingContainer::iterator p)
{
    modes.pushMode(*new WDeleteDrawingChartMode(p));
}

/************************** WMoveMarkerChartMode *************************/

WMoveMarkerChartMode::WMoveMarkerChartMode(GDrawingContainer::iterator drawing)
    : WMarkerChartMode(drawing)
{ }

void
WMoveMarkerChartMode::onMove(GChartLocation& loc)
{
    if (drawing->getPos() != loc.getLocation()) {
        drawing->setPos(loc.getLocation());
        loc.markChanged();
    }
    WMarkerChartMode::onMove(loc);
}

void
WMoveMarkerChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(COLOR_WHITE);

    GfxPoint sc = loc.getViewport().getCenter();

    string_t text = _(" Move marker, end with [ESC]. ");
    int wi = ctx.getFont().getTextWidth(text);
    int x = sc.x - wi / 2;
    int y = sc.y + 20;
    drawSolidBar(ctx, GfxRect(x, y, wi, font_heights[FONT_NORMAL]), COLOR_RED);
    outText(ctx, x, y, text);

    WMarkerChartMode::drawOverlays(can, loc);
}

bool
WMoveMarkerChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'v':
            // Exit move mode
            delete this;
            return true;
         case SDLK_F9:
         case SDLK_F9 + ss_Alt:
            // Change comment
            editMarkerComment(*drawing);
            loc.markChanged();
            return true;
        }
    }
    return WMarkerChartMode::handleEvent(event, loc);
}

/************************** WDrawMarkerChartMode *************************/

WDrawMarkerChartMode::WDrawMarkerChartMode(GDrawingContainer::iterator p)
    : WMarkerChartMode(p)
{ }

WDrawMarkerChartMode::~WDrawMarkerChartMode()
{
    finishDrawing();
}

void
WDrawMarkerChartMode::finishDrawing()
{
    GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
    if (drawing != c.end()) {
        if (drawing->getPos() == drawing->getPos2()) {
            c.erase(drawing);
            drawing = c.end();
        }
    }
}

void
WDrawMarkerChartMode::onMove(GChartLocation& loc)
{
    GPoint adj = loc.getWrapAdjust();
    if (adj.x || adj.y) {
        drawing->setPos(drawing->getPos() + adj);
        loc.markChanged();
    }
    if (drawing->getPos2() != loc.getLocation()) {
        drawing->setPos2(loc.getLocation());
        loc.markChanged();
    }
    WMarkerChartMode::onMove(loc);
}

void
WDrawMarkerChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(COLOR_WHITE);

    GfxPoint sc = loc.getViewport().view.getTopLeft();
    outText(ctx, sc.x, sc.y, _("Mark 2nd point, end with [ESC] or [+]."));

    WMarkerChartMode::drawOverlays(can, loc);
}

bool
WDrawMarkerChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case '+':
            if (drawing->getType() == LineDrawing) {
                // Line: start new line
                if (drawing->getPos() != drawing->getPos2()) {
                    GDrawing nd(*drawing);
                    nd.setPos(drawing->getPos2());
                    nd.setPos2(drawing->getPos2());
                    drawing = getDisplayedTurn().getCurrentUniverse().getDrawings().add(nd);
                    loc.markChanged();
                }
            } else {
                // Rectangle: exit move mode
                delete this;
            }
            return true;
         case '-':
            // No effect, because we handle '+' specially. Avoids that users
            // zoom out and cannot zoom in.
            return true;
         case SDLK_BACKSPACE:
            if (drawing->getPos() != drawing->getPos2()) {
                // Go to start point
                loc.setLocation(drawing->getPos());
            } else {
                // Trying to backspace from zero-size drawing, so exit mode
                delete this;
            }
            return true;
         case 'x':
            // Go to start point, swapping beginning and end of drawing
            {
                GPoint a = drawing->getPos(), b = drawing->getPos2();
                drawing->setPos(b);
                drawing->setPos2(a);
                loc.setLocation(a);
            }
            return true;
         case 'p':
            delete this;
            return true;
        }
    }
    return WMarkerChartMode::handleEvent(event, loc);
}

/************************** WDrawCircleChartMode *************************/

WDrawCircleChartMode::WDrawCircleChartMode(GDrawingContainer::iterator p)
    : WMarkerChartMode(p)
{ }

WDrawCircleChartMode::~WDrawCircleChartMode()
{ }

void
WDrawCircleChartMode::setRadius(int n)
{
    if (n > 0 && n < 1000 /* FIXME! */) {
        drawing->setCircleRadius(n);
        getParent()->markChanged();
    }
}

void
WDrawCircleChartMode::onMove(GChartLocation& loc)
{
    // Normally, user interface state should be checked in handleEvent(),
    // but this way it's much simpler.
    // FIXME: we should handle getWrapAdjust(). However, we shouldn't blindly
    // modify getCenter() and generate out-of-map circles.
    if (getCurrentModifierKeys() & ss_Alt)
        setRadius(roundToInt(std::sqrt(double(distanceSquared(loc.getCenter(), drawing->getPos())))));

    WMarkerChartMode::onMove(loc);
}

void
WDrawCircleChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(COLOR_WHITE);

    GfxPoint sc = loc.getViewport().view.getTopLeft();
    outText(ctx, sc.x, sc.y, _("Set circle size with [+]/[-], end with [ESC]."));
    outText(ctx, sc.x, sc.y + font_heights[FONT_NORMAL], format(_("Radius: %d ly"), drawing->getCircleRadius()));

    WMarkerChartMode::drawOverlays(can, loc);
}

bool
WDrawCircleChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case '+':
            setRadius(drawing->getCircleRadius() + 10);
            return true;
         case '-':
            setRadius(drawing->getCircleRadius() - 10);
            return true;
         case ss_Shift + '+':
            setRadius(drawing->getCircleRadius() + 1);
            return true;
         case ss_Shift + '-':
            setRadius(drawing->getCircleRadius() - 1);
            return true;
         case 'y':
            setRadius(350);
            return true;
         case SDLK_BACKSPACE:
            loc.setLocation(drawing->getPos());
            return true;
        }
    }

    return WMarkerChartMode::handleEvent(event, loc);
}

/************************ WDeleteDrawingChartMode ************************/

static void
deleteAdjacent(GDrawingContainer& c, GPoint pt)
{
    GDrawingContainer::iterator i = c.begin(), e = c.end();
    pt = pt.getNormalized();
    while (i != e) {
        if (i->isVisible() && i->getType() == LineDrawing) {
            if (i->getPos().getNormalized() == pt) {
                // Starts at given position
                pt = i->getPos2().getNormalized();
                c.erase(i);
                i = c.begin();
            } else if (i->getPos2().getNormalized() == pt) {
                // Ends at given position
                pt = i->getPos().getNormalized();
                c.erase(i);
                i = c.begin();
            } else {
                // No match
                ++i;
            }
        } else {
            ++i;
        }
    }
}

WDeleteDrawingChartMode::WDeleteDrawingChartMode(GDrawingContainer::iterator p)
    : WMarkerChartMode(p), counter(0), phase(false)
{ }

WDeleteDrawingChartMode::~WDeleteDrawingChartMode()
{ }

void
WDeleteDrawingChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    uint8 color = (phase ? COLOR_WHITE : COLOR_RED);
    loc.getViewport().drawDrawing(can, *drawing, color, true /* FIXME: use user setting instead? */);
    WMarkerChartMode::drawCursor(can, loc);
}

void
WDeleteDrawingChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    WMarkerChartMode::drawOverlays(can, loc);
    
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(COLOR_WHITE);

    GfxPoint sc = loc.getViewport().getCenter();
    string_t text = _(" Delete this drawing (y/n): ");
    int wi = ctx.getFont().getTextWidth(text);
    int x = sc.x - wi/2;
    int y = sc.y + 20;
    drawSolidBar(ctx, GfxRect(x, y, wi, font_heights[FONT_NORMAL]), COLOR_RED);
    outText(ctx, x, y, text);
    if (drawing->getType() == LineDrawing) {
        text = _(" Press [A] to delete adjacent lines, too. ");
        wi = ctx.getFont().getTextWidth(text);
        x = sc.x - wi/2;
        y += font_heights[FONT_NORMAL];
        drawSolidBar(ctx, GfxRect(x, y, wi, font_heights[FONT_NORMAL]), COLOR_RED);
        outText(ctx, x, y, text);
    }
}

bool
WDeleteDrawingChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    /* In PCC 1.x, Delete Mode eats all keyboard input except for
       (shifted) arrows, digits, and dot. Delete mode itself offers
       y (RET, SPC) and n (ESC), DEL to pick another object, and a
       to delete adjacent objects. */
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'y':
         case SDLK_RETURN:
         case ' ':
            // Delete marker
            {
                GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
                c.erase(drawing);
                drawing = c.end();
            }
            delete this;
            return true;

         case 'n':
         case SDLK_ESCAPE:
         case vk_Quit:
            // Do not delete
            delete this;
            return true;

         case SDLK_DELETE:
            // Let top-level handle this!
            // FIXME: should normally be parent who handles this
            return false;

         case 'a':
            // Delete adjacent
            if (drawing->getType() == LineDrawing) {
                GDrawingContainer& c = getDisplayedTurn().getCurrentUniverse().getDrawings();
                GPoint a = drawing->getPos(), b = drawing->getPos2();
                c.erase(drawing);
                drawing = c.end();
                deleteAdjacent(c, a);
                deleteAdjacent(c, b);
                delete this;
            }
            return true;

         default:
            {
                uint32 base_key = event.key.code & ~ss_AllMod;
                if (base_key == SDLK_LEFT || base_key == SDLK_RIGHT
                    || base_key == SDLK_UP || base_key == SDLK_DOWN
                    || base_key == vk_WheelUp || base_key == vk_WheelDown)
                {
                    // Handle these
                    return WMarkerChartMode::handleEvent(event, loc);
                } else {
                    // Swallow this key
                    return true;
                }
            }
        }
    }

    if (event.type == ev_Tick) {
        if (++counter >= TIMER_HZ/3) {
            counter = 0;
            phase = !phase;
            loc.markChanged();
        }
    }

    return WMarkerChartMode::handleEvent(event, loc);
}

void
editMarkerComment(GDrawing& marker)
{
    string_t comment = marker.getComment();
    if (UIInputLine::run(comment, 255 /* FIXME? */, _("Marker Comment"), _("Enter new comment for this marker:"), 30, 0))
        marker.setComment(strRTrim(comment));
}
