/**
  *  \file client/dlg-goals.cc
  *  \brief Build Goals Edit Dialog
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-goals.h"
#include "ui/window.h"
#include "ui/ui.h"
#include "ui/layout.h"
#include "util/nls.h"
#include "ui/widgets.h"
#include "gfx/palette.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/statictext.h"

namespace {
    class WEditBuildGoalsWindow : public UIWindow {
        GPlanet& planet;
        ObjectHolder h;

        UIDecimalSelector* input_widgets[2][4];

        UIBaseWidget& createInputWidget(int col, int line);
     public:
        WEditBuildGoalsWindow(GPlanet& planet);

        bool handleEvent(const UIEvent& e, bool second_pass);

        void writeBack();

        void setFocusToStructure(TPlanetaryStructure ps);
    };
}

// FIXME: this table also appears in dlg-structure.cc
static const char*const names[] = {
    N_("Mineral Mines:"),
    N_("Factories:"),
    N_("Defense Posts:"),
    N_("Starbase Defense:")
};

/** Constructor. Builds the complete window. */
WEditBuildGoalsWindow::WEditBuildGoalsWindow(GPlanet& planet)
    : UIWindow(_("Auto-Build Goals"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      planet(planet)
{
    // UIWindow [UIVBoxLayout]
    //   UIGroup [UIGridLayout]
    //     'Structure' 'Goal' 'Speed'
    //     'Mines'     Edit   Edit
    //     'Factories' Edit   Edit
    //     'Defense'   Edit   Edit
    //     'SBD'       Edit   Edit
    //   UIGroup [UIHBoxLayout]
    //     UIButton 'OK'
    //     UIButton 'Cancel'
    //     UISpacer
    //    [UIButton 'Help']
    //
    // Each Edit is:
    //     UIFocusGroup [UIHBoxLayout]
    //       UIButton '-'
    //       UIDecimalSelector
    //       UIButton '+'
    setLayout(UIVBoxLayout::instance5);

    UIGroup& group_grid    = h.add(new UIGroup(h.add(new UIGridLayout(3, 0))));
    UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));

    group_grid.add(h.add(new UIStaticText(_("Structure"), UIStaticText::StaticTitle)));
    group_grid.add(h.add(new UIStaticText(_("Goal"),      UIStaticText::StaticTitle)));
    group_grid.add(h.add(new UIStaticText(_("Speed"),     UIStaticText::StaticTitle)));

    for (int i = 0; i < 4; ++i) {
        group_grid.add(h.add(new UIStaticText(_(names[i]), UIStaticText::StaticTitle)));
        group_grid.add(createInputWidget(0, i));
        group_grid.add(createInputWidget(1, i));
        
        input_widgets[0][i]->setValue(planet.getAutobuildGoal(TPlanetaryStructure(i)));
        input_widgets[1][i]->setValue(planet.getAutobuildSpeed(TPlanetaryStructure(i)));
    }

    group_buttons.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    group_buttons.add(h.add(new UISpacer()));

    input_widgets[0][0]->setFocusedWidget();

    add(group_grid);
    add(group_buttons);
    pack();
}

/** Create one input widget. Input widgets consist of a UIFocusGroup containing
    the actual input widget (UIDecimalSelector), as well as buttons to control it.
    A pointer to the UIDecimalSelector will be stored in input_widgets[col][line].
    \param col Column (0=goal, 1=speed)
    \param line Line (structure type)
    \return reference to the whole widget group */
UIBaseWidget&
WEditBuildGoalsWindow::createInputWidget(int col, int line)
{
    const int widget_id   = 1000 + 2*line + col;
    const int lower_bound = 0;
    const int upper_bound = col==0 ? 1000 : 100;
    
    UIFocusGroup& g = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 0, 2));
    UIButton& minus = h.add(new UIButton("-", '-'));
    UIButton& plus  = h.add(new UIButton("+", '+'));
    UIDecimalSelector& ds = h.add(new UIDecimalSelector(widget_id, lower_bound, upper_bound, 10));

    g.add(minus);
    g.add(ds);
    g.add(plus);

    ds.setFocusedWidget();

    input_widgets[col][line] = &ds;

    return g;
}

/** Event handler. Handles focus hot-keys. */
bool
WEditBuildGoalsWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        /* what do we want to do? */
        int delta;
        if (e.key.code == SDLK_TAB)
            delta = 1;
        else if (e.key.code == SDLK_TAB + ss_Shift)
            delta = -1;
        else if (e.key.code == SDLK_DOWN)
            delta = 2;
        else if (e.key.code == SDLK_UP)
            delta = -2;
        else
            delta = 0;

        /* perform focus change */
        if (delta != 0) {
            UIBaseWidget* w = getFocusedWidget();
            if (w != 0) {
                int new_id = w->getId() + delta;
                if (new_id < 1000)
                    new_id += 8;
                else if (new_id >= 1008)
                    new_id -= 8;

                if (UIBaseWidget* new_widget = getWidgetById(new_id))
                    new_widget->setFocusedWidget();
            }
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

/** Write back changes to planet. */
void
WEditBuildGoalsWindow::writeBack()
{
    for (int i = 0; i < 4; ++i) {
        planet.setAutobuildGoal(TPlanetaryStructure(i), input_widgets[0][i]->getValue());
        planet.setAutobuildSpeed(TPlanetaryStructure(i), input_widgets[1][i]->getValue());
    }
}

/** Set user focus to a particular structure. */
void
WEditBuildGoalsWindow::setFocusToStructure(TPlanetaryStructure ps)
{
    input_widgets[0][ps]->setFocusedWidget();
}

/** Edit build goals on a planet. */
void
editBuildGoals(GPlanet& planet, TPlanetaryStructure initial_focus)
{
    WEditBuildGoalsWindow window(planet);
    window.setFocusToStructure(initial_focus);
    ui_root->centerWidget(window);
    ui_root->add(window);
    if (ui_root->run())
        window.writeBack();
    ui_root->close(window);
}
