/**
  *  \file client/dlg-list.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-list.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/msgbox.h"
#include "cpluslib/holder.h"
#include "client/dlg-visual.h"
#include "client/marks.h"

/************************** WObjectSelectionList *************************/

// FIXME: interaction with list.current() etc.

WObjectSelectionList::WObjectSelectionList(int id, int lines, int width, GObjectList& list)
    : UIListbox(id, lines, width, list.getNumObjects()), list(list),
      info_space(0),
      conn_univ_change(0),
      conn_turn_change(sig_turn_changed.add(this, &WObjectSelectionList::onTurnChange))
{
    // Register universe observer
    onTurnChange();

    // In case first object is not selectable, bring cursor onto selectable object
    scrollTo(0, 1);
}

WObjectSelectionList::~WObjectSelectionList()
{ }


/** Signal handler: turn change. Hook into new universe and redraw. */
void
WObjectSelectionList::onTurnChange()
{
    conn_univ_change = getDisplayedTurn().getCurrentUniverse().sig_pre_update.add(this, &WObjectSelectionList::onUniverseChange);
    drawWidget(false);
}

/** Signal handler: universe possibly changed. */
void
WObjectSelectionList::onUniverseChange()
{
    bool changed_any = false;
    bool changed_current = false;

    /* Check for changes */
    uint32_t index = getTop();
    for (int line = 0; line <= getPageSize(); ++line) {
        /* "line <= getPageSize()" to allow for partial last line */
        if (index < list.getNumObjects() && list.getObjectId(index+1).getObject().isDirty()) {
            if (index == getCurrent())
                changed_current = true;
            else
                changed_any = true;
        }
        ++index;
    }

    /* Apply changes */
    if (changed_any) {
        /* some objects changed */
        drawWidget(false);
    } else if (changed_current) {
        /* current object changed, we have a special update mechanism for that */
        updateCurrent();
    } else {
        /* no change */
    }
}

/** Fetch additional information for an object. Derived classes can override
    this to provide additional information for an object. For example,
    in proxy transfer, PCC will show current ship/ship transfers when
    selecting the proxy. */
string_t
WObjectSelectionList::getObjectInfo(GObjectId /*id*/)
{
    return string_t();
}

void
WObjectSelectionList::drawPart(GfxCanvas& can, int from, int to)
{
    const int h = getItemHeight();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);

    for (int line = from; line < to; ++line) {
        const int y = extent.y + h * line;
        const int index = getTop() + line;

        if (index >= list.getNumObjects()) {
            drawSolidBar(ctx, GfxRect(extent.x, y, extent.w, h), COLOR_GRAY);
        } else {
            GObjectId id = list.getObjectId(index+1);
            UIColor::Color scheme = UIColor::tc_Static;
            if (id.isGameObject())
                scheme = getPlayerColor(id.getObject().getOwner());
            int bgcolor, textcolor;
            if (index == getCurrent()) {
                // FIXME: rewrite using drawEntry
                textcolor = BLACK_COLOR_SET[scheme];
                bgcolor = getFocusState() != NoFocus ? COLOR_BLACK : COLOR_GRAY;
            } else {
                textcolor = GRAY_COLOR_SET[scheme];
                bgcolor = COLOR_GRAY;
            }
            if (id.getId()) {
                string_t info = getObjectInfo(id);
                int x = extent.x;
                int w = extent.w;

                drawSolidBar(ctx, GfxRect(x, y, w, h), bgcolor);
                if (id.getObject().isMarked() && w > info_space+15) {
                    // FIXME: selection color...
                    ctx.setColor(index == getCurrent() ? COLOR_YELLOW : COLOR_GREENBLACK);
                    drawSolidBar(ctx, GfxRect(x, y, 15, h), bgcolor);
                    drawSelection(ctx, GfxPoint(x + 9, y + h/2-1), 1, 2);
                    x += 15;
                    w -= 15;
                }

                ctx.setColor(textcolor);
                if (info.length()) {
                    outTextF(ctx, x, y, w - info_space, " " + id.getName());
                    outTextF(ctx, x + info_space, y, info_space, info);
                } else {
                    outTextF(ctx, x, y, w, " " + id.getName());
                }
            } else {
                /* divider line */
                ctx.setTransparentBackground();
                ctx.setColor(textcolor);
                drawSolidBar(ctx, GfxRect(extent.x, y, extent.w, h), COLOR_GRAY);
                outTextF(ctx, extent.x + 20, y, extent.w - 40, id.getName());
                ctx.setColor(COLOR_SHIELD+2);
                drawRectangle(ctx, GfxRect(extent.x, y, extent.w, h));
            }
        }
    }
}

bool
WObjectSelectionList::isAccessible(int index)
{
    return index >= 0
        && index < list.getNumObjects()
        && list.getObjectId(index+1).getId()
        && list.isExistant(index+1);
}

/** Select object /id/. \returns true iff ok, false if not. */
bool
WObjectSelectionList::selectObject(GObjectId id)
{
    int index = list.getIndexFor(id);
    if (index > 0) {
        scrollTo(index-1, 1);
        return true;
    }
    return false;
}

/** Process list change. Updates display according to new contents and also moves
    cursor to a specified object. */
void
WObjectSelectionList::onListChange(GObjectId objectToSelect)
{
    // Update bounds
    setSize(list.getNumObjects());

    // Go to object
    if (!selectObject(objectToSelect))
        scrollTo(0, 1);

    // Redraw everything
    drawWidget();
    sig_moved.raise(getId());
}

GObjectId
WObjectSelectionList::doStandardDialog(string_t title, int help_id)
{
    /* can we do something? */
    if (!list.getNumObjects())
        return GObjectId();

    /* okay, go ahead. */
    ObjectHolder holder;

    UIWindow win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(*this);

    UIGroup bgroup(UIHBoxLayout::instance5);
    bgroup.add(holder.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    bgroup.add(holder.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    bgroup.add(holder.add(new UISpacer()));
    if (help_id)
        /* FIXME */
        bgroup.add(holder.add(new UIButton(cm_Help, _("Help"), 'h')).addCommand());
    win.add(bgroup);
    win.setFocusedWidget(*this);
    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    if (rv)
        return getCurrentObject();
    else
        return GObjectId();
}
