/**
  *  \file client/dlg-subject.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-subject.h"
#include "ui/layout.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "util/misc.h"
#include "ui/widgets.h"

/*! \class WSubjectList
    \brief Message Subject List

    This widget displays a list of all message subjects of a
    GMessagebox. */

/** Constructor.
    \param id      Widget Id
    \param mbox    Mailbox box to display
    \param filter  Filter to apply to mailbox */
WSubjectList::WSubjectList(int id, GMessagebox& mbox, GMessageFilter& filter)
    : UIListbox(id, 20, 20 * font_ems[FONT_NORMAL], 0),
      mbox(mbox),
      filter(filter),
      groups()
{
    createGroupList();
}

WSubjectList::~WSubjectList()
{ }

/** Create the group list. */
void
WSubjectList::createGroupList()
{
    std::vector<Group> new_groups;

    int this_group_start = -1;
    string_t this_group_heading;

    for (int i = 0, max = mbox.getCount(); i < max; ++i) {
        string_t this_msg_heading = mbox.getHeading(i);
        if (this_msg_heading != this_group_heading) {
            if (this_group_start >= 0)
                new_groups.push_back(Group(this_group_start, i - this_group_start));
            this_group_heading = this_msg_heading;
            this_group_start = i;
        }
    }
    if (this_group_start >= 0)
        new_groups.push_back(Group(this_group_start, mbox.getCount() - this_group_start));

    new_groups.swap(groups);
    setSize(groups.size());

    // FIXME: redraw, reposition
}

/** Get selected message. Since each row can represent multiple
    messages, this returns the first message of the row. */
int
WSubjectList::getSelectedMessage() const
{
    if (getCurrent() < groups.size()) {
        return groups[getCurrent()].first_msg;
    } else {
        return 0;
    }
}

/** Select a message. Selects the row which contains the message. */
void
WSubjectList::selectMessage(int n)
{
    for (std::vector<Group>::size_type i = 0; i < groups.size(); ++i) {
        if (groups[i].first_msg >= n && n - groups[i].first_msg < groups[i].count) {
            scrollTo(i, 1);
            break;
        }
    }
}

bool
WSubjectList::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Focused) && e.type == ev_Key) {
        if (e.key.code == 'k') {
            // Toggle kill
            if (getCurrent() < groups.size()) {
                filter.toggleFilter(mbox.getHeading(groups[getCurrent()].first_msg));
                drawWidget();
            }
            return true;
        }
    }
    return UIListbox::handleEvent(e, second_pass);
}

void
WSubjectList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        const int index = getTop() + i;
        if (index >= groups.size()) {
            drawEntry(can, UIColor::tc_Static, i, "");
        } else {
            // draw by hand. FIXME: this is a copy of drawEntry.
            const int real_y = getItemHeight()*i + extent.y + getHeaderHeight();
            const int divi = 3*font_ems[FONT_NORMAL];
            UISkin& s = getSkin();

            const UIColorSet& scheme = (index != getCurrent() ? GRAY_COLOR_SET : getFocusState() != NoFocus ? BLACK_COLOR_SET : BRIGHT_GRAY_COLOR_SET);
            drawSolidBar(ctx, GfxRect(extent.x, real_y, extent.w, getItemHeight()), scheme[UIColor::tc_Background]);
            ctx.setColor(scheme[UIColor::tc_Static]);
            if (filter.isFiltered(mbox.getHeading(groups[index].first_msg)))
                ctx.setColor(scheme[UIColor::tc_Faded]);

            ctx.setTextAlign(2, 0);
            outTextF(ctx, extent.x + divi, real_y, divi, numToString(groups[index].count));
            ctx.setTextAlign(0, 0);
            outTextF(ctx, extent.x + divi, real_y, extent.w - divi,
                     " " + mbox.getHeading(groups[index].first_msg));
        }
    }
}

/** Do a standard dialog.
    \return nonzero if user chose OK */
int
WSubjectList::doStandardDialog()
{
    UIWindow window(_("Message Subjects"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    window.setLayout(UIVBoxLayout::instance5);

    UIGroup group_buttons(UIHBoxLayout::instance5);
    UIButton button_ok(1, _("OK"), SDLK_RETURN);
    UIButton button_cancel(0, _("Cancel"), SDLK_ESCAPE);
    UISpacer spacer;
    button_ok.addStop();
    button_cancel.addStop();

    group_buttons.add(button_ok);
    group_buttons.add(button_cancel);
    group_buttons.add(spacer);

    window.add(*this);
    window.add(group_buttons);
    window.pack();

    setFocusedWidget();

    ui_root->centerWidget(window);
    ui_root->add(window);
    int result = ui_root->run();
    ui_root->close(window);

    return result;
}
