/**
  *  \file client/scr-ship.cc
  *  \brief Starship Control Screen
  *
  *  This contains the ship screen tile widgets, and the main driver
  *  code.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/actions.h"
#include "client/cscreen.h"
#include "client/dialogs/consoledlg.h"
#include "client/dlg-cargo.h"
#include "client/dlg-list.h"
#include "client/dlg-mission.h"
#include "client/dlg-objsel.h"
#include "client/dlg-warp.h"
#include "client/remote.h"
#include "client/tiles/shipoverview.h"
#include "client/tiles/shipmission.h"     // cm_Ship_BaseMenu
#include "client/wscanchart.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/host.h"
#include "game/planet.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/spec.h"
#include "game/team.h"
#include "game/univ.h"
#include "gfx/clip.h"
#include "gfx/palette.h"
#include "int/data.h"
#include "int/if/shipif.h"
#include "int/intglobal.h"
#include "int/keymap.h"
#include "resmgr/resid.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "ui/widgets/framegroup.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/tiles.h"

namespace {
    /// Ship scanner (with waypoint etc).
    class WShipScannerChartWidget : public WScannerChartWidget {
     protected:
        void  lockQueryLocation(bool& hyperjumping, int& current_gravity_pid);
        int32 lockQueryDistance(GPoint pt);
        
     public:
        WShipScannerChartWidget();
        ~WShipScannerChartWidget();
        void drawScanner(GfxCanvas& can);
        void drawPost(GfxCanvas& can);
        // void update(int);
        void userMoveScannerTo(GPoint pt);
    };

    /// Starship screen.
    class WShipScreen : public WControlScreen {
        WShipScannerChartWidget scanner_chart;

        void doShipBaseMenu(GShip& ship, UIBaseWidget& button);
     public:
        WShipScreen();
        ~WShipScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void update();
        void setScriptContext();
        void setWaypointToScanner();
    };
}

static int acquireWaypointPerm(const string_t& reason, GShip& ship);

int
chooseShip()
{
    int nships = global_ship_type.getCount();
    if (nships == 0) {
        messageBox(_("You do not have any ships. "
                     "To build ships, use [B] on a starbase."),
                   _("Starships"));
        return 0;
    } else if (nships == 1) {
        return global_ship_type.getCurrent();
    } else {
        GAliasObjectType aot(global_ship_type);
        WShipOverviewTile bot(aot);
        WObjectSelectionDialog dlg(_("Select Ship"), aot);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter)
            return aot.getCurrent();
        else
            return 0;
    }
}

/************************ WShipScannerChartWidget ************************/

WShipScannerChartWidget::WShipScannerChartWidget()
{ }

WShipScannerChartWidget::~WShipScannerChartWidget()
{ }

void
WShipScannerChartWidget::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    const GShip& sh = global_ship_type.getShip();
    hyperjumping        = sh.isHyperdriving();
    current_gravity_pid = getDisplayedTurn().getCurrentUniverse().getGravityPlanetAt(sh.getPos());
}

int32
WShipScannerChartWidget::lockQueryDistance(GPoint pt)
{
    const GShip& sh = global_ship_type.getShip();
    const int32 dist2 = distanceSquared(sh.getPos(), pt);
    if (sh.isHyperdriving() && !host.isExactHyperjumpDistance2(dist2)) {
        return -1;
    } else {
        return dist2;
    }
}

void
WShipScannerChartWidget::drawScanner(GfxCanvas& can)
{
    const GChartViewport& vp = getViewport();

    GPoint                wp = global_ship_type.getShip().getWaypoint();
    GfxPoint       wp_screen = vp.scaleXY(wp);

    if (wp != getScannerPosition() && wp_screen != vp.getCenter()) {
        GfxContext ctx;
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.useCanvas(can);
        ctx.setColor(COLOR_RED);
        drawLine(ctx, vp.getCenter(), wp_screen);
        drawEndBig(ctx, wp_screen);
    }

    WScannerChartWidget::drawScanner(can);
}

void
WShipScannerChartWidget::drawPost(GfxCanvas& can)
{
    // Drawing parameters
    GfxContext ctx;
    const GChartViewport& vp = getViewport();
    GfxClipRectFilter can1(getExtent(), can);

    ctx.useCanvas(can1);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.setTextAlign(1, 0);

    // Draw mine laying info
    GMineMissionData md;
    if (parseMineLayingMission(global_ship_type.getShip(), getDisplayedTurn().getCurrentUniverse(), md)) {
        // We're laying a minefield. What kind?
        GMinefield* mf = 0;
        if (md.mine_id > 0)
            mf = getDisplayedTurn().getCurrentUniverse().getMinefields().get(md.mine_id);

        // Figure out where it goes
        GPoint center;
        int radius;
        string_t notice;
        if (mf == 0) {
            // New minefield
            center = global_ship_type.getShip().getPos();
            radius = GMinefield::getRadiusFromUnits(md.num_units);
            notice = format(_("%d ly"), radius);
        } else {
            // Enlarge existing minefield
            center = mf->getPos();
            radius = GMinefield::getRadiusFromUnits(md.num_units + mf->getUnitsForLaying());
            notice = format(_("%d ly (+%d)"), radius, radius - mf->getRadius());
        }

        GfxPoint p = vp.scaleXY(center.getNearestTo(vp.center));
        ctx.setColor(COLOR_YELLOW);

        // Limit check
        int limit = (md.type
                     ? config.MaximumWebMinefieldRadius(md.owner)
                     : config.MaximumMinefieldRadius(md.owner));
        if (radius > limit) {
            drawCircle(ctx, p, vp.scale(limit));
            ctx.setColor(COLOR_DARK);
            outText(ctx, p.x, p.y + 10 + font_heights[FONT_SMALL], _("<over limit>"));
        }

        // Draw circle
        drawCircle(ctx, p, vp.scale(radius));
        outText(ctx, p.x, p.y + 10, notice);
    } else if (parseMineScoopMission(global_ship_type.getShip(), getDisplayedTurn().getCurrentUniverse(), md)) {
        // We're scooping
        const GMinefieldContainer& mfc = getDisplayedTurn().getCurrentUniverse().getMinefields();
        const GShip& sh = global_ship_type.getShip();
        int room = sh.getFreeCargo();
        for (GMinefieldContainer::const_iterator mf = mfc.begin(); mf != mfc.end(); ++mf) {
            // Check whether we can scoop this field
            if (mf->getOwner() == md.owner
                && distanceSquared(mf->getPos(), sh.getPos()) <= mf->getUnitsForLaying()
                && (md.mine_id == 0 || md.mine_id == mf->getId()))
            {
                // Okay, scoop it. First, figure out conversion rate.
                int rate = ((mf->isWeb() ? config.UnitsPerWebRate(md.owner) : config.UnitsPerTorpRate(md.owner))
                            * square(sh.getTorpType()))
                    / 100;
                if (rate == 0)
                    rate = 1;

                // How much do we clear?
                int32 units = mf->getUnitsForLaying();
                int32 scoopedUnits;
                int scoopedTorps = units / rate;
                if (scoopedTorps > room) {
                    // We cannot clear this field
                    scoopedTorps = room;
                    scoopedUnits = room * rate;
                } else {
                    // We can clear this field
                    scoopedUnits = units;
                }

                // Check limit
                if (md.num_torps > 0 && md.num_torps < scoopedTorps) {
                    scoopedTorps = md.num_torps;
                    scoopedUnits = md.num_torps * rate;
                }

                // Draw it
                if (scoopedTorps > 0) {
                    GfxPoint p = vp.scaleXY(mf->getPos().getNearestTo(vp.center));
                    ctx.setColor(COLOR_YELLOW);
                
                    room -= scoopedTorps;
                    if (units > scoopedUnits) {
                        // Field stays
                        int radius = mf->getRadiusFromUnits(units - scoopedUnits);
                        drawCircle(ctx, p, radius);
                        outText(ctx, p.x, p.y + 10, format(_("%d ly (%d torp%!1{s%})"), radius, scoopedTorps));
                    } else {
                        // Field is gone
                        outText(ctx, p.x, p.y + 10, format(_("gone (%d torp%!1{s%})"), scoopedTorps));
                    }
                }
            }
        }
    }

    // Base class
    WScannerChartWidget::drawPost(can);
}

// void
// WShipScannerChartWidget::update(int)
// {
//     drawWidget(false);
// }

// void
// WShipScannerChartWidget::handleScannerChange(bool dblclick)
// {
//     if (dblclick) {
//         GGameTurn& trn = getDisplayedTurn();
//         if (int id = acquireWaypointPerm(_("waypoint"), trn.getShipType().getShip())) {
//             trn.getShip(id).setWaypoint(getScannerPosition());
//             trn.getCurrentUniverse().doUpdates();
//         }
//     }
// }

void
WShipScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    bool scanner_tied_to_waypoint = false;
    if (scanner_tied_to_waypoint) {
        // FIXME
        setScannerPosition(pt);
    } else {
        setScannerPosition(pt);
    }
}

/****************************** WShipScreen ******************************/

WShipScreen::WShipScreen()
    : WControlScreen(global_ship_type, WScreenId::ShipScreen),
      scanner_chart()
{
    createTileLayout(*this, "SHIPSCREEN", global_ship_type);
    addChart(scanner_chart);
    setKeymap(IntKeymap::getKeymapByName("SHIPSCREEN"));

    scanner_chart.sig_dblclick.add(this, &WShipScreen::setWaypointToScanner);
}

WShipScreen::~WShipScreen()
{ }

void
WShipScreen::update()
{
    int sid = getType().getCurrent();
    const GShip& ship = getType().getUniverse()->getShip(sid);
    scanner_chart.setCenterPosition(ship.getPos());
    scanner_chart.setScannerPosition(scanner_chart.clipScannerPosition(ship.getWaypoint()));
    scanner_chart.drawWidget();
}

void
WShipScreen::setScriptContext()
{
    global_contexts.push_back(new IntShipContext(getType().getCurrent()));
}

/** Attempt to acquire permission to change waypoint. (this is used for fleets)
    \param reason reason, for user
    \param ship   ship we want to change
    \returns 0 if permission not granted, ship id to change waypoint on
    specified ship */
static int
acquireWaypointPerm(const string_t& reason, GShip& ship)
{
    if (!ship.getWaypointPermission()) {
        // FIXME: display dialog, leave fleet, ...
        (void) reason;
        return 0;
    } else {
        return ship.getId();
    }
}

void
doShipTransferFor(GShip& ship, GObjectId chosen)
{
    GMultiTransaction mt;
    GShipTransfer&   left = mt.addNewSubtransaction(new GShipTransfer(ship));
    GCargoContainer* right;

    if (!chosen.isGameObject()) {
        /* jettison */
        right = &mt.addNewSubtransaction(new GShipTransporter(ship));
    } else if (GShip* s = dynamic_cast<GShip*>(&chosen.getObject())) {
        if (s->getOwner() == ship.getOwner()) {
            /* ship <-> ship */
            ASSERT(s->hasShipData());
            right = &mt.addNewSubtransaction(new GShipTransfer(*s));
        } else {
            /* ship <-> foreign ship */
            if (ship.isTransporterInUse(GShip::t_Transfer)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            if (ship.getTransporter(GShip::t_Transfer).id != 0
                && ship.getTransporter(GShip::t_Transfer).id != s->getId()
                && ship.isTransporterActive(GShip::t_Transfer))
            {
                messageBox(format(_("The cargo transporter is already transferring cargo "
                                    "to ship #%d."),
                                  ship.getTransporter(GShip::t_Transfer).id), _("Cargo Transfer"));
                // FIXME: state name of current target ship here
                // FIXME: we could cancel the transfer here
                return;
            }
            // ship.cancelTranporter(true);
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Transfer, *s));
        }
    } else if (GPlanet* p = dynamic_cast<GPlanet*>(&chosen.getObject())) {
        if (p->getOwner() == ship.getOwner() && p->hasPlanetData()) {
            right = &mt.addNewSubtransaction(new GPlanetTransfer(*p));
        } else {
            if (ship.isTransporterInUse(GShip::t_Unload)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Unload, *p));
        }
    } else {
        FAIL();
        right = 0;
    }

    doCargoTransfer(mt, mt.addNewSubtransaction(new GCargoTransfer(left, *right)));
}

static void
doShipCargoTransfer(GUniverse& univ, GShip& ship)
{
    GObjectList list;
    int pid = univ.getPlanetAt(ship.getPos());
    if (pid)
        list.addObject(global_anyplanet_type, pid);
    else
        list.addNonObject(-1, _("Jettison into space"));

    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (i != ship.getId() && univ.getShip(i).isExistant() && univ.getShip(i).getPos() == ship.getPos())
            list.addObject(global_anyship_type, i);

    GObjectId chosen = WObjectSelectionList(0, 15, 320, list).doStandardDialog(_("Transfer cargo to..."), 0 /* FIXME */);
    if (chosen.isValid())
        doShipTransferFor(ship, chosen);
}

static void
doShipNavigationChart(GUniverse& univ, GShip& ship)
{
    int size = ui_root->getExtent().h;

    UIWindow the_window(_("Navigation Chart"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    the_window.setLayout(UIVBoxLayout::instance5);

    ObjectHolder h;

    WScannerChartWidget chart;
    chart.setCenterPosition(ship.getPos());
    chart.setScannerPosition(ship.getWaypoint());
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    UIFrameGroup& fg = h.add(new UIFrameGroup(UIVBoxLayout::instance0));
    fg.add(chart);
    the_window.add(fg);

    UIGroup buttons(UIHBoxLayout::instance5);
    WScanResult scanner;
    scanner.setChart(chart);
    buttons.add(scanner);
    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10)).addStop());
    buttons.add(h.add(new UIButton(cm_Escape, "ESC",        SDLK_ESCAPE)).addStop());
    the_window.add(buttons);
    the_window.setExtent(GfxRect(0, 0, size - 50, size - 10));

    ui_root->centerWidget(the_window);
    ui_root->add(the_window);
    int rv = ui_root->run();
    ui_root->close(the_window);

    if (rv == cm_Enter) {
        ship.setWaypoint(chart.getScannerPosition());
        setOptimumWarp(univ, ship.getId());
    }
}

void
WShipScreen::setWaypointToScanner()
{
    GGameTurn& trn = getDisplayedTurn();
    if (int id = acquireWaypointPerm(_("waypoint"), global_ship_type.getShip())) {
        trn.getCurrentUniverse().getShip(id).setWaypoint(getScannerPosition());
        setOptimumWarp(trn.getCurrentUniverse(), id);
        trn.getCurrentUniverse().doScreenUpdates();
    }
}

void
WShipScreen::doShipBaseMenu(GShip& ship, UIBaseWidget& button)
{
    enum {
        CancelShipyard,
        FixThisShip,
        RecycleThisShip,
        GoToBase
    };

    // Are we orbiting a base?
    if (!global_base_type.isExistant(ship.getOrbit())) {
        messageBox(_("We are not at one of our starbases."), _("Starbase Commands"));
        return;
    }
    GPlanet& base = global_base_type.getObject(ship.getOrbit());
    GUniverse& univ = *global_base_type.getUniverse();

    // Build command menu
    UIStringList cmdlist;

    // - Fix/Recycle
    // FIXME: the "can we fix" logic is also in act-fix.cc; can we share it?
    int currentAction = base.getBaseShipyardId() == ship.getId() ? base.getBaseShipyardOrder() : 0;
    if (currentAction == sya_Fix) {
        cmdlist.add(CancelShipyard, _("Cancel \"repair\" order"));
    } else {
        cmdlist.add(FixThisShip, _("Fix (repair) this ship"));
    }
    if (currentAction == sya_Recycle) {
        cmdlist.add(CancelShipyard, _("Cancel \"recycle\" order"));
    } else {
        if (base.getOwner() == ship.getOwner())
            cmdlist.add(RecycleThisShip, _("Recycle this ship"));
    }

    // - Build torps/fighters FIXME

    // - Starbase screen
    cmdlist.add(GoToBase, _("Starbase screen"));

    // - Clone FIXME

    // Show menu
    UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], cmdlist) + 20, cmdlist);
    if (!lb.doStandardMenu(button))
        return;

    switch (int cmd = lb.getKey()) {
     case CancelShipyard:
        base.setBaseShipyardOrder(0, 0);
        univ.doScreenUpdates();
        break;

     case FixThisShip:
     case RecycleThisShip:
        {
            bool accept = true;
            if (base.getBaseShipyardOrder() != sya_None && base.getBaseShipyardId() != ship.getId()) {
                // Find name of ship being repaired/recycled
                string_t shipName;
                if (univ.isValidShipId(base.getBaseShipyardId()) && univ.getShip(base.getBaseShipyardId()).isExistant())
                    shipName = univ.getShip(base.getBaseShipyardId()).getName();
                else
                    shipName = format(_("Ship %d"), base.getBaseShipyardId());
                // Ask user
                if (base.getBaseShipyardOrder() == sya_Fix) {
                    accept = UIMessageBox(format(_("This starbase is already repairing %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                } else if (base.getBaseShipyardOrder() == sya_Recycle) {
                    accept = UIMessageBox(format(_("This starbase is already recycling %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                }
            }
            if (accept) {
                base.setBaseShipyardOrder(cmd == FixThisShip ? sya_Fix : sya_Recycle,
                                          ship.getId());
                univ.doScreenUpdates();
            }
        }
        break;

     case GoToBase:
        if (ship.getOrbit() > 0)
            doGoToScreen(3, ship.getOrbit());
        break;
    }
}

bool
WShipScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        GUniverse& univ = *getType().getUniverse();
        GShip&     ship = global_ship_type.getShip();
        switch (e.key.code) {
            // 'n', 'e', 'm' implemented in core.q
         case 'a':
            if (int id = acquireWaypointPerm(_("waypoint"), ship)) {
                doShipNavigationChart(univ, univ.getShip(id));
                univ.doScreenUpdates();
            }
            return true;
         case 'c':
            doShipCargoTransfer(univ, ship);
            univ.doScreenUpdates();
            return true;
         case ss_Ctrl + 'j':
            if (univ.getPlanetAt(ship.getPos()) == 0) {
                doShipTransferFor(ship, GObjectId());
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 's':
            if (ship.getTransporter(GShip::t_Transfer).id) {
                doShipTransferFor(ship, GObjectId(global_anyship_type,
                                                  ship.getTransporter(GShip::t_Transfer).id));
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 'p':
            // FIXME: should possibly also handle beam-up orders
            if (ship.getTransporter(GShip::t_Unload).id) {
                doShipTransferFor(ship, GObjectId(global_anyplanet_type,
                                                  ship.getOrbit()));
                univ.doScreenUpdates();
            }
            return true;
         case 'u':
            if (ship.getOrbit()) {
                doShipTransferFor(ship, GObjectId(global_anyplanet_type,
                                                  ship.getOrbit()));
                univ.doScreenUpdates();
            } else {
                messageBox(_("We are not orbiting a planet, sir. "
                             "If you want to jettison cargo, use the "
                             "'c' command instead."),
                           _("Cargo Transfer"));
            }
            return true;
         case 'w':
            if (int id = acquireWaypointPerm(_("speed"), ship)) {
                (WShipSpeedSelector(univ.getShip(id)).doStandardDialog());
            }
            return true;
         case 'r':
            if (host.isPHost() && config.CPEnableRemote())
                doRemoteControl(ship, getDisplayedTurn().getCommands(ship.getOwner()));
            univ.doScreenUpdates();
            return true;
         case ss_Ctrl + 'z':
            if (int id = acquireWaypointPerm(_("waypoint"), ship)) {
                ship.setWaypoint(ship.getPos());
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 'w':
            setWaypointToScanner();
            return true;
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
         case ss_Ctrl + SDLK_TAB:
         case ss_Shift + ss_Ctrl + SDLK_TAB: {
            GObjectList l;
            l.addObjectsAt(univ, ship.getPos(), 0, 0);
            l.postprocessUser(false);
            int cur = l.getIndexFor(ship);
            int n = ((e.key.code & ss_Shift)
                     ? l.getPreviousWrap(cur, (e.key.code & ss_Ctrl) != 0)
                     : l.getNextWrap    (cur, (e.key.code & ss_Ctrl) != 0));
            if (n > 0)
                getType().setCurrent(l.getObject(n).getId());
            return true;
         }

            /* ship screen keys:

               m-i, m-t         // search xceptor/tower
               c-n              // list next turn
               q                // toggle independantmove
               d                // distribute cargo
               f10              // fleet
               s-f10            // join fleet/goto fleet
               m-r              // reset x/y
            */
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

bool
WShipScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_Ship_BaseMenu) {
        doShipBaseMenu(global_ship_type.getShip(), sender);
        return true;
    } else {
        return WControlScreen::handleCommand(cmd, sender);
    }
}

void
doShipScreen()
{
    WShipScreen().runScreen();
}
