/**
  *  \file game/load.h
  *  \brief Loading Functions
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_LOAD_H
#define PCC_V2_GAME_LOAD_H

#include "cpluslib/smartptr.h"
#include "cpluslib/types.h"
#include "io/dirs.h"
#include "io/streams.h"
#include "game/struct.h"
#include "cpluslib/assert.h"
#include "game/playerset.h"

class GGameTurn;

/** \defgroup game_load Game Loading Functions.
    These functions load game data into a GGameTurn. @{ */

/** Target for a Load operation. Note that these must have these exact
    values, as the loadXXX functions will do 'where & LoadCurrent'. */
enum LoadUniverse {
    LoadCurrent  = 1,
    LoadPrevious = 2,
    LoadBoth     = LoadCurrent | LoadPrevious
};

/** Target file format. */
enum TargetFormat {
    TargetPlaintext,            ///< Plaintext file. Standard in Dosplan etc.
    TargetEncrypted             ///< Encrypted file. Winplan's additional targets.
};

class GDirectoryOverview {
    int pp_flags[NUM_PLAYERS];
    int global_flags;
 public:
    GDirectoryOverview();

    /** Per-player flags. Note that most combinations are possible; see class description. */
    enum {
        have_RST        = 1,       ///< We have a current RST.
        have_TRN        = 2,       ///< We have a current TRN.
        have_Gamedir    = 4,       ///< We have a current game directory.
        have_NewRST     = 8,       ///< We have a new RST (along with game dir).
        have_Conflict   = 16
    };
    void scan(Directory& dir, bool rstOnly = false);
    void clear();

    /** Get flags (have_XXX) for one player. */
    int getPlayerFlags(int pid) const
        { ASSERT(pid > 0 && pid <= NUM_PLAYERS); return pp_flags[pid-1]; }
    /** Get global flags. Global flags are the bitwise OR of all
        player flags. */
    int getGlobalFlags() const
        { return global_flags; }

    GPlayerSet getPlayersWhere(int flags) const;
    int getDefaultPlayer() const;
};

/** Logic to read an outbox file. */
class GOutboxReader {
 public:
    /** Add message read from file.
        \param message Message text, decoded
        \param receivers Receiver set */
    virtual void addMessage(string_t message, GPlayerSet receivers) = 0;

    void loadOutbox(Stream& s);
    void loadOutbox35(Stream& s);
};

/* single-load routines, load.cc */
void loadShips(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src);
void loadPlanets(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src);
void loadBases(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src);
void loadTargets(GGameTurn& trn, Stream& s, int count, TargetFormat fmt, LoadUniverse where, GPlayerSet src);
void loadShipXY(GGameTurn& trn, Stream& s, int avail, LoadUniverse where, GPlayerSet reject);
void loadKoreMinefields(GGameTurn& trn, Stream& s, int count);
void loadKoreIonStorms(GGameTurn& trn, Stream& s, int count);
void loadUfos(GGameTurn& trn, Stream& s, int first_id, int count);
void loadKoreExplosions(GGameTurn& trn, Stream& s, int count);
void addMessageFromFile(GGameTurn& trn, string_t text, int sender, GPlayerSet recv);

/* complete routines, load-{rst,trn,dir}.cc */
void loadResult(GGameTurn& trn, Stream& s, int player);
void loadTurn(GGameTurn& trn, Stream& s, int player);
void loadDirectory(GGameTurn& trn, Directory& d, int player);
void loadHostDirectory(GGameTurn& trn, Directory& d);

/*@}*/

#endif
