/**
  *  \file int/binary.cc
  *  \brief Interpreter: Binary Operations
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/binary.h"
#include "cpluslib/misc.h"
#include "int/value.h"
#include "int/interror.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "cpluslib/assert.h"
#include "int/keymap.h"

/********************************* Naming ********************************/

static const char*const binary_names[] = {
    "and",
    "or",
    "xor",
    "add",
    "sub",
    "mul",
    "div",
    "idiv",
    "rem",
    "pow",
    "concat",
    "concempty",
    "cmpeq",
    "cmpeqNC",
    "cmpne",
    "cmpneNC",
    "cmple",
    "cmpleNC",
    "cmplt",
    "cmpltNC",
    "cmpge",
    "cmpgeNC",
    "cmpgt",
    "cmpgtNC",
    "min",
    "minNC",
    "max",
    "maxNC",
    "firststr",
    "firststrNC",
    "reststr",
    "reststrNC",
    "findstr",
    "findstrNC",
    "bitand",
    "bitor",
    "bitxor",
    "str",
    "atan",
    "lcut",
    "rcut",
    "endcut",
    "strmult",
    "keyaddparent",
    "keyfind",
};

/** Get name for a binary operation. This is used for disassembling.
    \param op Operation, IntBinaryOperation */
const char*
getBinaryName(uint8_t op)
{
    if (op < countof(binary_names))
        return binary_names[op];
    else
        return "?";
}

/******************************* Execution *******************************/

namespace {
    /** Comparison result. */
    enum Comparison {
        cmpNull    = 1,         ///< Either argument is null.
        cmpLess    = 2,         ///< a < b.
        cmpEqual   = 4,         ///< a = b.
        cmpGreater = 8          ///< a > b.
    };

    /** Arithmetic status. */
    enum Arithmetic {
        ariBad,                 ///< Bad argument combination.
        ariNull,                ///< Either argument is null.
        ariInt,                 ///< Use integer arithmetic.
        ariFloat                ///< Use float arithmetic.
    };

    /** Arithmetic argument pair. */
    struct ArithmeticPair {
        int32_t ia, ib;         ///< Operands for integer arithmetic.
        double fa, fb;          ///< Operands for float arithmetic.
    };
}

/** Check arguments for arithmetic.
    \param pair [out] Receives the operands to operate on
    \param a,b  [in] User-supplied operands
    \return argument status */
static Arithmetic
checkArithmetic(ArithmeticPair& pair, IntValue* a, IntValue* b)
{
    bool afloat = false, bfloat = false;

    // Check for null
    if (a == 0 || b == 0)
        return ariNull;

    // Check a for numericness
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(a))
        pair.ia = iv->getValue();
    else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(a))
        pair.fa = fv->getValue(), afloat = true;
    else
        return ariBad;

    // Check b for numericness
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(b))
        pair.ib = iv->getValue();
    else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(b))
        pair.fb = fv->getValue(), bfloat = true;
    else
        return ariBad;

    // Promote if necessary
    if (afloat || bfloat) {
        if (!afloat)
            pair.fa = pair.ia;
        if (!bfloat)
            pair.fb = pair.ib;
        return ariFloat;
    } else {
        return ariInt;
    }
}

static string_t
convertCase(IntStringValue* sv, bool doit)
{
    if (doit)
        return strUCase(sv->getValue());
    else
        return sv->getValue();
}

/** Perform comparison. Compares two user-supplied parameters and returns
    comparison result. Caller converts this again to a user-visible value.
    \param a,b User-supplied parameters
    \return Result */
static Comparison
compare(IntValue* a, IntValue* b, bool caseblind)
{
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return cmpNull;
     case ariInt:
        return p.ia < p.ib
            ? cmpLess
            : p.ia == p.ib
            ? cmpEqual
            : cmpGreater;
     case ariFloat:
        return p.fa < p.fb
            ? cmpLess
            : p.fa == p.fb
            ? cmpEqual
            : cmpGreater;
     default:
        IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
        IntStringValue* sb = dynamic_cast<IntStringValue*>(b);
        if (sa != 0 && sb != 0) {
            string_t ssa = convertCase(sa, caseblind);
            string_t ssb = convertCase(sb, caseblind);
            return ssa < ssb ? cmpLess : ssa == ssb ? cmpEqual : cmpGreater;
        }
        throw IntError::typeError();
    }
}

static IntValue*
FAnd(IntValue* a, IntValue* b)
{
    // Logical And, ternary logic
    //   e_f_t
    // e|e f e
    // f|f f f
    // t|e f t
    int ba = getBoolValue(a);
    int bb = getBoolValue(b);
    if (ba > 0 && bb > 0)
        return makeBoolValue(1);
    else if (ba == 0 || bb == 0)
        return makeBoolValue(0);
    else
        return makeBoolValue(-1);
}

static IntValue*
FOr(IntValue* a, IntValue* b)
{
    // Logical Or, ternary logic
    //   e_f_t
    // e|e e t
    // f|e f t
    // t|t t t
    int ba = getBoolValue(a);
    int bb = getBoolValue(b);
    if (ba > 0 || bb > 0)
        return makeBoolValue(1);
    else if (ba == 0 && bb == 0)
        return makeBoolValue(0);
    else
        return makeBoolValue(-1);
}

static IntValue*
FXor(IntValue* a, IntValue* b)
{
    // Logical Xor, ternary logic
    //   e_f_t
    // e|e e e
    // f|e f t
    // t|e t f
    int ba = getBoolValue(a);
    int bb = getBoolValue(b);
    if (ba < 0 || bb < 0)
        return makeBoolValue(-1);
    else
        return makeBoolValue(ba ^ bb);
}

static IntValue*
FAdd(IntValue* a, IntValue* b)
{
    // Arithmetical addition or string concatenation
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(p.ia + p.ib);
     case ariFloat:
        return makeFloatValue(p.fa + p.fb);
     default:
        IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
        IntStringValue* sb = dynamic_cast<IntStringValue*>(b);
        if (sa != 0 && sb != 0)
            return makeStringValue(sa->getValue() + sb->getValue());
        throw IntError::typeError();
    }
}

static IntValue*
FSub(IntValue* a, IntValue* b)
{
    // Subtraction
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(p.ia - p.ib);
     case ariFloat:
        return makeFloatValue(p.fa - p.fb);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FMult(IntValue* a, IntValue* b)
{
    // Multiplication
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(p.ia * p.ib);
     case ariFloat:
        return makeFloatValue(p.fa * p.fb);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FDivide(IntValue* a, IntValue* b)
{
    // Division
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        // FIXME: PCC1.x returns integer if result is integer
        if (p.ib == 0)
            throw IntError("Divide by zero");
        return makeFloatValue(double(p.ia) / double(p.ib));
     case ariFloat:
        if (std::fabs(p.fb) < 1.0E-06)          // FIXME?
            throw IntError("Divide by zero");
        return makeFloatValue(p.fa / p.fb);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FIntegerDivide(IntValue* a, IntValue* b)
{
    // Integer division
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        if (p.ib == 0)
            throw IntError("Divide by zero");
        return makeIntValue(p.ia / p.ib);
     default:
        throw IntError::typeError(IntError::tyInteger);
    }
}

static IntValue*
FRemainder(IntValue* a, IntValue* b)
{
    // Integer remainder
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        if (p.ib == 0)
            throw IntError("Divide by zero");
        return makeIntValue(p.ia % p.ib);
     default:
        throw IntError::typeError(IntError::tyInteger);
    }
}

static IntValue*
FPow(IntValue* a, IntValue* b)
{
    // Exponentiation
    if (a == 0 || b == 0)
        return 0;

    /* Second argument must be integer */
    IntIntValue* bi = dynamic_cast<IntIntValue*>(b);
    if (!bi)
        throw IntError::typeError(IntError::tyInteger);

    /* First argument must be integer or real */
    if (IntIntValue* ai = dynamic_cast<IntIntValue*>(a)) {
        /* FIXME: precision */
        double result = std::pow(double(ai->getValue()), double(bi->getValue()));
        if (bi->getValue() >= 0 && result >= -2147483648.0 && result < 2147483648.0) {
            /* FIXME: without this, 61^2 yields 3720. 
               A correct fix would be to compute completely in integers. */
            if (result > 0)
                result += 0.00001;
            else
                result -= 0.00001;
            return makeIntValue(int32_t(result));
        } else {
            return makeFloatValue(result);
        }
    } else if (IntFloatValue* af = dynamic_cast<IntFloatValue*>(a)) {
        return makeFloatValue(std::pow(af->getValue(), bi->getValue()));
    } else {
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FConcat(IntValue* a, IntValue* b)
{
    // Concatenation, null annihilates
    if (a == 0 || b == 0)
        return 0;
    else
        return makeStringValue(a->toString(false) + b->toString(false));
}

static IntValue*
FConcatEmpty(IntValue* a, IntValue* b)
{
    // Concatenation, null interpolates
    if (a == 0 && b == 0) {
        // FIXME: PCC 1.x does not special-case this, and returns "" for EMPTY & EMPTY.
        return 0;
    } else {
        string_t result;
        if (a != 0)
            result += a->toString(false);
        if (b != 0)
            result += b->toString(false);
        return makeStringValue(result);
    }
}

/****************** Optionally case-sensitive functions ******************/

namespace {

    IntValue*
    FCompareEQ(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compare for equality
        Comparison cmp = compare(a, b, caseblind);
        return (cmp & cmpNull)
            ? 0
            : makeBoolValue((cmp & cmpEqual) != 0);
    }

    IntValue*
    FCompareNE(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compare for inequality
        Comparison cmp = compare(a, b, caseblind);
        return (cmp & cmpNull)
            ? 0
            : makeBoolValue((cmp & (cmpGreater | cmpLess)) != 0);
    }

    IntValue*
    FCompareLE(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compare for less/equal
        Comparison cmp = compare(a, b, caseblind);
        return (cmp & cmpNull)
            ? 0
            : makeBoolValue((cmp & (cmpLess | cmpEqual)) != 0);
    }

    IntValue*
    FCompareLT(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compare for less than
        Comparison cmp = compare(a, b, caseblind);
        return (cmp & cmpNull)
            ? 0
            : makeBoolValue((cmp & cmpLess) != 0);
    }

    IntValue*
    FCompareGE(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compare for greater/equal
        Comparison cmp = compare(a, b, caseblind);
        return (cmp & cmpNull)
            ? 0
            : makeBoolValue((cmp & (cmpGreater | cmpEqual)) != 0);
    }

    IntValue*
    FCompareGT(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compare for greater than
        Comparison cmp = compare(a, b, caseblind);
        return (cmp & cmpNull)
            ? 0
            : makeBoolValue((cmp & cmpGreater) != 0);
    }

    IntValue*
    FMin(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compute minimum
        Comparison cmp = compare(a, b, caseblind);
        if (cmp & cmpNull)
            return 0;
        else if (cmp & cmpLess)
            return a->clone();
        else
            return b->clone();
    }

    IntValue*
    FMax(IntValue* a, IntValue* b, bool caseblind)
    {
        // Compute maximum
        Comparison cmp = compare(a, b, caseblind);
        if (cmp & cmpNull)
            return 0;
        else if (cmp & cmpGreater)
            return a->clone();
        else
            return b->clone();
    }

    IntValue*
    FFirstStr(IntValue* a, IntValue* b, bool caseblind)
    {
        // Split string at delimiter, return first part
        if (a == 0 || b == 0)
            return 0;

        IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
        IntStringValue* sb = dynamic_cast<IntStringValue*>(b);
        if (sa == 0 || sb == 0)
            throw IntError::typeError(IntError::tyString);

        string_t ssa = sa->getValue();
        string_t::size_type apos = convertCase(sa, caseblind).find(convertCase(sb, caseblind));
        if (apos != string_t::npos)
            ssa.erase(apos);

        return makeStringValue(ssa);
    }

    IntValue*
    FRestStr(IntValue* a, IntValue* b, bool caseblind)
    {
        // Split string at delimiter, return remainder
        if (a == 0 || b == 0)
            return 0;

        IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
        IntStringValue* sb = dynamic_cast<IntStringValue*>(b);
        if (sa == 0 || sb == 0)
            throw IntError::typeError(IntError::tyString);

        string_t::size_type apos = convertCase(sa, caseblind).find(convertCase(sb, caseblind));
        if (apos != string_t::npos) {
            return makeStringValue(sa->getValue().substr(apos + sb->getValue().size()));
        } else {
            return 0;
        }
    }

    IntValue*
    FFindStr(IntValue* a, IntValue* b, bool caseblind)
    {
        // Find substring
        if (a == 0 || b == 0)
            return 0;

        IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
        IntStringValue* sb = dynamic_cast<IntStringValue*>(b);
        if (sa == 0 || sb == 0)
            throw IntError::typeError(IntError::tyString);

        string_t::size_type apos = convertCase(sa, caseblind).find(convertCase(sb, caseblind));
        if (apos != string_t::npos)
            return makeIntValue(apos+1);
        else
            return makeIntValue(0);
    }

}

static IntValue*
FBitAnd(IntValue* a, IntValue* b)
{
    // Bitwise and
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(p.ia & p.ib);
     default:
        throw IntError::typeError(IntError::tyInteger);
    }
}

static IntValue*
FBitOr(IntValue* a, IntValue* b)
{
    // Bitwise or
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(p.ia | p.ib);
     default:
        throw IntError::typeError(IntError::tyInteger);
    }
}

static IntValue*
FBitXor(IntValue* a, IntValue* b)
{
    // Bitwise Xor
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(p.ia ^ p.ib);
     default:
        throw IntError::typeError(IntError::tyInteger);
    }
}

static IntValue*
FStr(IntValue* a, IntValue* b)
{
    // Stringification with precision
    if (a == 0 || b == 0)
        return 0;

    /* Check second arg */
    IntIntValue* bi = dynamic_cast<IntIntValue*>(b);
    if (bi == 0)
        throw IntError::typeError(IntError::tyInteger);
    if (bi->getValue() < 0)
        throw IntError::rangeError();

    /* Check first arg */
    if (IntIntValue* ai = dynamic_cast<IntIntValue*>(a)) {
        /* Bool converts as-is, as does integer with precision 0 */
        if (ai->isBool() || bi->getValue() == 0)
            return makeStringValue(ai->toString(false));
        /* Convert integer as floating point */
        return makeStringValue(format(string_t(format("%%.%df", bi->getValue())).c_str(), double(ai->getValue())));
    } else if (IntFloatValue* af = dynamic_cast<IntFloatValue*>(a)) {
        /* Convert float */
        return makeStringValue(format(string_t(format("%%.%df", bi->getValue())).c_str(), double(af->getValue())));
    } else {
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FAtan(IntValue* a, IntValue* b)
{
    // Arc-tangent
    double value;
    ArithmeticPair p;
    switch (checkArithmetic(p, a, b)) {
     case ariNull:
        return 0;
     case ariInt:
        p.fa = p.ia;
        p.fb = p.ib;
        /* FALLTHROUGH */
     case ariFloat:
        if (p.fa == 0 && p.fb == 0)
            return 0;
        value = std::atan2(p.fa, p.fb) * (180.0 / PI);
        if (value < 0)
            value += 360;
        return makeFloatValue(value);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FLCut(IntValue* a, IntValue* b)
{
    // Remove leftmost N characters
    if (a == 0 || b == 0)
        return 0;

    IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
    IntIntValue*    ib = dynamic_cast<IntIntValue*>(b);
    if (sa == 0 || ib == 0)
        throw IntError::typeError();

    string_t ssa = sa->getValue();
    int32_t  iib = ib->getValue();
    if (iib > 0) {
        --iib;
        if (uint32_t(iib) < ssa.size())
            ssa.erase(0, iib);
        else
            ssa.clear();
    }
    return makeStringValue(ssa);
}

static IntValue*
FRCut(IntValue* a, IntValue* b)
{
    // Remove after Nth character
    if (a == 0 || b == 0)
        return 0;

    IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
    IntIntValue*    ib = dynamic_cast<IntIntValue*>(b);
    if (sa == 0 || ib == 0)
        throw IntError::typeError();

    string_t ssa = sa->getValue();
    int32_t  iib = ib->getValue();
    if (iib > 0) {
        if (uint32_t(iib) < ssa.size())
            ssa.erase(iib);
    } else {
        ssa.clear();
    }
    return makeStringValue(ssa);
}

static IntValue*
FEndCut(IntValue* a, IntValue* b)
{
    // Remove all but last N characters
    if (a == 0 || b == 0)
        return 0;

    IntStringValue* sa = dynamic_cast<IntStringValue*>(a);
    IntIntValue*    ib = dynamic_cast<IntIntValue*>(b);
    if (sa == 0 || ib == 0)
        throw IntError::typeError();

    string_t ssa = sa->getValue();
    int32_t  iib = ib->getValue();
    if (iib > 0) {
        if (uint32_t(iib) < ssa.size())
            ssa.erase(0, ssa.size() - iib);
    } else {
        ssa.clear();
    }
    return makeStringValue(ssa);
}

static IntValue*
FStrMult(IntValue* a, IntValue* b)
{
    // Replicate string
    if (a == 0 || b == 0)
        return 0;

    IntIntValue*    ia = dynamic_cast<IntIntValue*>(a);
    IntStringValue* sb = dynamic_cast<IntStringValue*>(b);
    if (ia == 0 || sb == 0)
        throw IntError::typeError();

    int32_t  iia = ia->getValue();
    string_t ssb = sb->getValue();

    // FIXME: we should have some kind of limits to avoid overloading
    string_t result;
    while (iia > 0) {
        result += ssb;
        --iia;
    }

    return makeStringValue(result);
}

static IntValue*
FKeyAddParent(IntValue* a, IntValue* b)
{
    if (a == 0 || b == 0)
        return 0;

    IntKeymapValue* keymap = dynamic_cast<IntKeymapValue*>(a);
    IntKeymapValue* parent = dynamic_cast<IntKeymapValue*>(b);
    if (keymap == 0 || parent == 0)
        throw IntError::typeError(IntError::tyKeymap);

    keymap->getKeymap()->addParent(parent->getKeymap());
    return keymap->clone();
}

static IntValue*
FKeyFind(IntValue* a, IntValue* b)
{
    if (a == 0 || b == 0)
        return 0;

    // Keymap
    IntKeymapValue* keymap = dynamic_cast<IntKeymapValue*>(a);
    if (keymap == 0)
        throw IntError::typeError(IntError::tyKeymap);

    // Key
    IntStringValue* keysym = dynamic_cast<IntStringValue*>(b);
    if (!keysym)
        throw IntError::typeError(IntError::tyString);
    uint32_t keyval = parseKey(keysym->getValue());
    if (!keyval)
        throw IntError("Invalid key name");

    // Generate result
    uint32_t cmd = keymap->getKeymap()->lookupCommand(keyval);
    if (!cmd)
        return 0;
    else
        return makeIntValue(cmd);
}

template<IntValue* (*Func)(IntValue*, IntValue*, bool), bool Value>
static IntValue*
Bind(IntValue* a, IntValue* b)
{
    return Func(a, b, Value);
}

static IntValue* (*const binary_ops[])(IntValue*,IntValue*) = {
    FAnd,
    FOr,
    FXor,
    FAdd,
    FSub,
    FMult,
    FDivide,
    FIntegerDivide,
    FRemainder,
    FPow,
    FConcat,
    FConcatEmpty,
    Bind<FCompareEQ,false>,
    Bind<FCompareEQ,true>,
    Bind<FCompareNE,false>,
    Bind<FCompareNE,true>,
    Bind<FCompareLE,false>,
    Bind<FCompareLE,true>,
    Bind<FCompareLT,false>,
    Bind<FCompareLT,true>,
    Bind<FCompareGE,false>,
    Bind<FCompareGE,true>,
    Bind<FCompareGT,false>,
    Bind<FCompareGT,true>,
    Bind<FMin,false>,
    Bind<FMin,true>,
    Bind<FMax,false>,
    Bind<FMax,true>,
    Bind<FFirstStr,false>,
    Bind<FFirstStr,true>,
    Bind<FRestStr,false>,
    Bind<FRestStr,true>,
    Bind<FFindStr,false>,
    Bind<FFindStr,true>,
    FBitAnd,
    FBitOr,
    FBitXor,
    FStr,
    FAtan,
    FLCut,
    FRCut,
    FEndCut,
    FStrMult,
    FKeyAddParent,
    FKeyFind,
};

CTASSERT(countof(binary_ops) == countof(binary_names));

/** Execute binary operation.
    \param op Operation, IntBinaryOperation
    \param a,b User-supplied arguments taken from value stack
    \return New value to push on value stack */
IntValue*
executeBinaryOperation(uint8_t op, IntValue* a, IntValue* b)
{
    if (op < countof(binary_ops))
        return binary_ops[op](a, b);
    else
        throw IntError::internalError("invalid binary operation");
}
