#!/usr/bin/perl -w
#
#  perl int/names/namelist.pl [/path/to/source] [< namesX.txt]
#
#  Reads the name lists for PCC1 (if path given) or PCC2 (default).
#  Merges this with the comments from namesX.txt, if given.
#  Outputs a new version of namesX.txt to stdout.
#
#  Items which appear in the text file but not in the source code will
#  be marked as UNDEFINED source.
#
use strict;

# name -> kind -> { implemented=>[list], note=>[list] }
my %names;

# Find source directory
my $srcdir = shift @ARGV;
if (!defined($srcdir)) {
    foreach (qw(. .. ../..)) {
        if (-r "$_/version.h") {
            $srcdir = $_;
            last;
        }
    }
}

if (-r "$srcdir/cc.pas") {
    #
    #  PCC 1.x
    #

    # Some hardcoded values
    foreach (qw(AND OR XOR NOT TRUE FALSE PI MOD)) {
        $names{$_}{"fund-op"}{implemented} = ["ccexpr.pas"]
    }
    $names{COMMENT}{"planet-prop"}{implemented} = ["core.q"];
    $names{COMMENT}{"ship-prop"}{implemented} = ["core.q"];
    foreach (qw(STORAGE.ENGINES STORAGE.HULLS STORAGE.BEAMS STORAGE.LAUNCHERS STORAGE.AMMO)) {
        $names{$_}{"planet-prop"}{implemented} = ["planint.pas"];
    }

    # Fundamental functions
    readFundamentalFunctions("$srcdir/ccexpr.pas");

    # Name lists
    readCommandInfo("$srcdir/ccexec.pas",  "global-cmd", "fund-stmt");
    readCommandInfo("$srcdir/globint.pas", "global-cmd", "fund-stmt");
    readCommandInfo("$srcdir/planint.pas", "planet-cmd");
    readCommandInfo("$srcdir/shipint.pas", "ship-cmd");
    # missing: values.pas, minefield, marker

    # IDL Name maps
    readIdlNameMap("$srcdir/ship.sym",   "AllShipsMapReader",     "ship-prop");
    # TargetMapReader is a subset of OwnShipMapReader
    readIdlNameMap("$srcdir/ship.sym",   "OwnShipMapReader",      "ship-prop");
    readIdlNameMap("$srcdir/ship.sym",   "ShipHullspecMapReader", "ship-prop");
    readIdlNameMap("$srcdir/ship.sym",   "HullspecMapReader",     "hull-prop");
    readIdlNameMap("$srcdir/ship.sym",   "ShipHullspecMapReader", "hull-prop");
    readIdlNameMap("$srcdir/global.sym", "GlobalAlwaysMapReader", "global-prop");
    readIdlNameMap("$srcdir/global.sym", "GlobalPlayerMapReader", "global-prop");
    readIdlNameMap("$srcdir/planet.sym", "AllPlanetsMapReader",   "planet-prop");
    readIdlNameMap("$srcdir/planet.sym", "PlanetMapReader",       "planet-prop");
    readIdlNameMap("$srcdir/planet.sym", "BaseMapReader",         "planet-prop");
    readIdlNameMap("$srcdir/values.sym", "PlayerMapReader",       "player-prop");
    readIdlNameMap("$srcdir/values.sym", "AllPlayerMapReader",    "player-prop");
    readIdlNameMap("$srcdir/values.sym", "MinefieldMapReader",    "mine-prop");
    readIdlNameMap("$srcdir/values.sym", "UfoMapReader",          "ufo-prop");
    readIdlNameMap("$srcdir/values.sym", "IonMapReader",          "ion-prop");
    readIdlNameMap("$srcdir/values.sym", "VcrSideMapReader",      "marker-prop", "LEFT.");
    readIdlNameMap("$srcdir/values.sym", "VcrSideMapReader",      "marker-prop", "RIGHT.");
    readIdlNameMap("$srcdir/values.sym", "VcrAllMapReader",       "vcr-prop");
    readIdlNameMap("$srcdir/values.sym", "MarkerMapReader",       "marker-prop");
    readIdlNameMap("$srcdir/specs.sym",  "EngineMapReader",       "engine-prop");
    readIdlNameMap("$srcdir/specs.sym",  "TorpedoMapReader",      "torp-prop");
    # LauncherMapReader is the same as TorpedoMapReader
    readIdlNameMap("$srcdir/specs.sym",  "BeamMapReader",         "beam-prop");
    # BuildOrderMapReader is not implemented in PCC2
    # AutoTaxMapReader is not implemented anywhere

    # Regular definitions
    readScriptDefinition("$srcdir/core.q");
} else {
    #
    #  PCC2
    #

    # Fundamental operators are hardcoded
    foreach (qw(AND OR XOR NOT TRUE FALSE PI MOD)) {
        $names{$_}{"fund-op"}{implemented} = ["lex.cc"]
    }

    # Name lists
    readNameList("$srcdir/int/keywords.cc",      "keyword_table",     "fund-stmt");
    readNameList("$srcdir/int/builtin.cc",       "builtin_functions", "fund-func");
    readNameList("$srcdir/int/if/planetif.cc",   "planet_mapping",    "planet-prop", "planet-cmd");
    readNameList("$srcdir/int/if/shipif.cc",     "ship_mapping",      "ship-prop",   "ship-cmd");
    readNameList("$srcdir/int/if/specif.cc",     "beam_map",          "beam-prop",   "beam-cmd");
    readNameList("$srcdir/int/if/specif.cc",     "engine_map",        "engine-prop", "engine-cmd");
    readNameList("$srcdir/int/if/specif.cc",     "torpedo_map",       "torp-prop",   "torp-cmd");
    readNameList("$srcdir/int/if/hullif.cc",     "hull_mapping",      "hull-prop",   "hull-cmd");
    readNameList("$srcdir/int/if/ionif.cc",      "ion_mapping",       "ion-prop",    "ion-cmd");
    readNameList("$srcdir/int/if/ufoif.cc",      "ufo_mapping",       "ufo-prop",    "ufo-cmd");
    readNameList("$srcdir/int/if/mineif.cc",     "minefield_mapping", "mine-prop",   "mine-cmd");
    readNameList("$srcdir/int/if/playerif.cc",   "player_mapping",    "player-prop", "player-cmd");
    readNameList("$srcdir/int/if/globalif.cc",   "global_mapping",    "global-prop", "global-cmd");

    # "Manual" definitions
    readGlobalDefinition("$srcdir/int/if/guiif.cc",  "global_property_names", "global-var", "global");
    readGlobalDefinition("$srcdir/int/if/conif.cc",  "global_property_names", "global-var", "global");
    readGlobalDefinition("$srcdir/int/if/initif.cc", "global_property_names", "global-var", "global");
    readGlobalDefinition("$srcdir/int/intglobal.cc", "global_property_names", "global-var", "global");
    readGlobalDefinition("$srcdir/int/intglobal.cc", "ship_property_names",   "ship-prop");
    readGlobalDefinition("$srcdir/int/intglobal.cc", "planet_property_names", "planet-prop");

    # Regular definitions
    readScriptDefinition("$srcdir/resource/core.q");
}

# Read existing file
readExistingDefinitions()
  unless -t STDIN;

printf "%-30s  %-14s  %-22s  %s\n", "Name", "Kind", "Implemented", "Note";
print "------------------------------  --------------  ----------------------  ----------------------\n";

foreach my $name (sort keys %names) {
    foreach my $kind (sort keys %{$names{$name}}) {
        printf "%-30s  %-14s  %-22s  %s\n",
          $name, $kind,
            join(", ", sort @{$names{$name}{$kind}{implemented}}),
              exists($names{$name}{$kind}{note}) ? join(", ", sort @{$names{$name}{$kind}{note}}) : "";
    }
}

# Read Name List
# Looks in $file for a table named $variable containing one name per line,
# and generates entries of type $kind. If $kind2 is specified, this kind
# is used when the respective line contains the keyword Method, i.e. to
# distinguish between
#    { "BUILDTORPS",                21,                   PlanetMethod },
#    { "COLONISTS",                 ippColonists,         PlanetProperty },
sub readNameList {
    my $file = shift;
    my $variable = shift;
    my $kind = shift;
    my $kind2 = shift;

    my $fileBaseName = $file;
    $fileBaseName =~ s|.*/||;

    my $did = 0;

    open FILE, "< $file" or die "$file: $!";
    while (<FILE>) {
        if (/^[a-zA-Z].*\b$variable\b.*\{/) {
            # This starts our variable
            while (<FILE>) {
                s|//.*||;
                if (/^\}/) {
                    # end
                    last
                } elsif (/"(.*)"/) {
                    # it's a variable
                    my $name = $1;
                    my $thisKind = $kind;
                    if (defined($kind2)) {
                        if (/Method/) { $thisKind = $kind2; }
                    }
                    push @{$names{$name}{$thisKind}{implemented}}, $fileBaseName;
                    $did = 1;
                }
            }
            last;
        }
    }
    close FILE;
    if (!$did) {
        die "$file did not contain any items of type $kind\n";
    }
}

# Read global definition.
# Looks in $file for commands that manipulate a property name variable $props
#     global_values.setNew(global_property_names.addMaybe(NAME), new ....);
#     global_property_names.add("SYSTEM.ERR");
#     planet_property_names.add("COMMENT");
#     defineGlobalValue("NAME", value);
#     defineGlobalProcedure("NAME", value);
# and adds these as type $kind-prop, $kind-func, $kind-cmd.
sub readGlobalDefinition {
    my $file = shift;
    my $props = shift;
    my $kind = shift;
    my $kind2 = shift;

    my $fileBaseName = $file;
    $fileBaseName =~ s|.*/||;

    my $did = 0;

    open FILE, "< $file" or die "$file: $!";
    while (<FILE>) {
        if (/global_values\.setNew\($props\.add(Maybe)?\("(.*?)"\), *(.*?)\)/) {
            # First form
            my $name = $2;
            my $init = $3;
            my $thisKind = defined($kind2)
              ? ($init =~ /IntSimpleProcedureValue/ ? "$kind2-cmd"
                 : $init =~ /IntSimpleIndexableValue/ ? "$kind2-func"
                 : "$kind2-var")
                : $kind;
            push @{$names{$name}{$thisKind}{implemented}}, $fileBaseName;
            $did = 1;
        } elsif (/defineGlobalValue\("(.*?)", *(.*?)\)/) {
            # Fourth form
            my $name = $1;
            my $init = $2;
            my $thisKind = defined($kind2)
              ? ($init =~ /IntSimpleProcedureValue/ ? "$kind2-cmd"
                 : $init =~ /IntSimpleIndexableValue/ ? "$kind2-func"
                 : "$kind2-var")
                : $kind;
            push @{$names{$name}{$thisKind}{implemented}}, $fileBaseName;
            $did = 1;
        } elsif (/defineGlobalProcedure\("(.*?)",/) {
            # Fifth form
            my $name = $1;
            push @{$names{$name}{defined($kind2) ? "$kind2-cmd" : "$kind"}{implemented}}, $fileBaseName;
            $did = 1;
        } elsif (/$props\.add(Maybe)?\("(.*?)"\)/) {
            # Second/third form
            my $name = $2;
            push @{$names{$name}{"$kind"}{implemented}}, $fileBaseName;
            $did = 1;
        }
    }
    close FILE;
    if (!$did) {
        die "$file did not contain any items of type $kind\n";
    }
}

# Read script definitions
sub readScriptDefinition {
    my $file = shift;

    my $fileBaseName = $file;
    $fileBaseName =~ s|.*/||;

    my $did = 0;

    open FILE, "< $file" or die "$file: $!";
    while (<FILE>) {
        s|%.*||;
        if (/\b(sub|function) +([a-z0-9\$_.]+)/i) {
            # Function
            my $name = uc($2);
            my $thisKind = lc($1) eq 'sub' ? "global-cmd" : "global-func";
            push @{$names{$name}{$thisKind}{implemented}}, $fileBaseName
              unless $name =~ m|^CC\$|;
            $did = 1;
        } elsif (/\b(dim\s+|shared\s+)+([a-z0-9\$_.]+)/i) {
            # Variable
            my $name = uc($2);
            push @{$names{$name}{"global-var"}{implemented}}, $fileBaseName
              unless $name =~ m|^CC\$|;
            $did = 1;
        }
    }
    close FILE;
    if (!$did) {
        die "$file did not contain any items\n";
    }
}


# Read command info structures (PCC v1)
sub readCommandInfo {
    my $file = shift;
    my $kind = shift;
    my $kind2 = shift;

    my $fileBaseName = $file;
    $fileBaseName =~ s|.*/||;

    my $did = 0;

    # First gather all elements in %allCommands, because there will be duplicates
    my %allCommands;
    open FILE, "< $file" or die "$file: $!";
    while (<FILE>) {
        if (m|\(id: *CMD_([A-Z_0-9]+); *argspec.*exec|) {
            my $name = $1;
            my $thisKind = $kind;
            if (defined($kind2) && /ci_Special/) { $thisKind = $kind2 }
            $allCommands{$name} = $thisKind;
            $did = 1;
        }
    }
    close FILE;
    if (!$did) {
        die "$file did not contain any items\n";
    }

    foreach my $k (keys %allCommands) {
        unless ($k =~ /^CC_/) {
            push @{$names{convertName($k)}{$allCommands{$k}}{implemented}}, $fileBaseName;
        }
    }
}

# Read PCCv1 IDL name mappings
sub readIdlNameMap {
    my $file = shift;
    my $var = shift;
    my $kind = shift;
    my $prefix = shift || "";

    my $fileBaseName = $file;
    $fileBaseName =~ s|.*/||;

    my $did = 0;
    open FILE, "< $file" or die "$file: $!";
    while (<FILE>) {
        if (/\b$var\b *: *array.*tnamemap *= *\(/i) {
            while (<FILE>) {
                if (/^\s*\)/) {
                    # end
                    last
                } elsif (/\(id:\s*Ofs\(__ID_([A-Z0-9_]+)\);\s*writer.*Typ/) {
                    push @{$names{convertName($prefix.$1)}{$kind}{implemented}}, $fileBaseName;
                    $did = 1;
                }
            }
            last;
        }
    }

    close FILE;
    if (!$did) {
        die "$file did not contain any items for $kind\n";
    }
}

# Read PCC1 fundamental function list
sub readFundamentalFunctions {
    my $file = shift;

    my $fileBaseName = $file;
    $fileBaseName =~ s|.*/||;

    my $did = 0;
    open FILE, "< $file" or die "$file: $!";
    my $content = join('', <FILE>);
    while ($content =~ m/func_names *: *array.*?of\s*char\s*=\s*(.*);/sig) {
        my $ele = $1;
        while ($ele =~ m/'([A-Z0-9_.\$]+)'/g) {
            push @{$names{$1}{"fund-func"}{implemented}}, $fileBaseName;
        }
    }
    close FILE;
}


# Read existing definitions from standard input
sub readExistingDefinitions {
    my $reading = 0;
    my $regexp;
    while (<STDIN>) {
        chomp;
        s/\r//;
        if (!$reading) {
            if (/^([- ]+)/) {
                $reading = 1;
            }
        } else {
            my @content = split /  +/;
            if (@content < 3 || $content[0] =~ /[a-z]/ || $content[1] =~ /[A-Z]/) {
                print STDERR "Warning: line not understood:\n$_";
            } else {
                my $name = $content[0];
                my $kind = $content[1];
                my @implemented = split /, */, $content[2];
                my @note = @content > 3 ? split /, */, $content[3] : ();

                if (!exists $names{$name}{$kind}) {
                    $names{$name}{$kind} = { implemented => ["UNDEFINED"],
                                             note => [@note] };
                } else {
                    $names{$name}{$kind}{note} = [@note];
                }
            }
        }
    }
}

sub convertName {
    my $name = shift;
    $name =~ s/_/./g;
    $name =~ s/\.$/\$/;
    $name =~ s/^CC\./CC\$/;
    $name;
}
