/**
  *  \file io/textfile.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "io/textfile.h"
#include "io/streams.h"
#include "cpluslib/assert.h"
#include "cpluslib/types.h"

/** Constructor.
    \param s Stream to read */
TextFile::TextFile(Stream& s)
    : stream(s),
      line_nr(0),
      buf_ptr(0),
      buf_fill(0),
      eof(false)
{ }

/** Destructor. */
TextFile::~TextFile()
{ }

/** Refill buffer. */
void
TextFile::refillBuffer()
{
    ASSERT(buf_ptr >= buf_fill);
    if (!eof) {
        buf_ptr = 0;
        buf_fill = stream.read(buffer, sizeof(buffer));
        if (buf_fill == 0)
            eof = true;
    }
}

/** Read a line from the text file. Accepts all possible newline conventions.
    \param line [out] Line
    \return true if line read, false if EOF reached */
bool
TextFile::getLine(string_t& line)
{
    // Read line
    line.clear();
    int ch;
    while ((ch = getByte()) >= 0 && (ch != 10 && ch != 13))
        line.append(1, char(ch));

    // If CRLF, read the LF
    if (ch == 13 && peekByte() == 10)
        getByte();

    // If the file ends in Ctrl-Z, remove that
    if (ch < 0 && line.size() && line[line.size()-1] == 26)
        line.erase(line.size()-1);

    if (ch >= 0 || line.size()) {
        ++line_nr;
        return true;
    } else {
        return false;
    }
}

/** Read a byte from the text file.
    \return character read (>= 0), or -1 on EOF */
int
TextFile::getByte()
{
    if (buf_ptr >= buf_fill)
        refillBuffer();
    return eof ? -1 : uint8(buffer[buf_ptr++]);
}

/** Preview a byte from the text file without reading it. The byte is
    not removed from the stream until you call getByte().
    \return character read (>= 0), or -1 on EOF */
int
TextFile::peekByte()
{
    if (buf_ptr >= buf_fill)
        refillBuffer();
    return eof ? -1 : uint8(buffer[buf_ptr]);
}
