/**
  *  \file resmgr/resdir.cc
  *  \brief ResDirectory resource provider class
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "resmgr/resdir.h"
#include "gfx/font.h"
#include "io/textfile.h"
#include "util/console.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"        // FIXME
#include "gfx/blit.h"

/** Convert resource identifier string value into file name template.
    For example, "foo.bar90" is converted to "foo/bar90.", which tells
    openResourceFile() to look for a file "bar90" with any extension
    within directory "foo".
    \param name [in/out] Resource name */
static void
convertName(string_t& name)
{
    string_t::size_type n = 0;
    while ((n = name.find('.', n)) != string_t::npos)
        name[n] = Platform::PATH_SEPARATOR, ++n;
    name += '.';
}

/****************************** ResDirectory *****************************/

/*! \class ResDirectory

    This manages resources stored in individual files. It provides
    management of alias definitions (index.txt) as well as heuristics
    for guessing file names from resource Ids: a resource "foo.bar" is
    looked up in file "foo/bar.EXT", where EXT is an extension
    matching the resource type.

    For graphics, it supports some additional tricks. */

/** Constructor. */
ResDirectory::ResDirectory(Ptr<Directory> dir)
    : dir(dir)
{
    Ptr<Stream> s = dir->openFileNT("index.txt", Stream::C_READ);
    if (s) {
        loadAliases(*s);
    }
}

ResDirectory::~ResDirectory()
{ }

Ptr<GfxPixmap>
ResDirectory::loadPixmap(ResId id)
{
    /* Fetch definition for this item */
    string_t name = id.toString();
    alias_map_t::iterator i = alias_map.find("g." + name);
    if (i != alias_map.end()) {
        name = i->second.name;
    } else {
        convertName(name);
    }

    /* Parse it */
    string_t file = strRTrim(strFirst(name, "|"));
    Ptr<GfxPixmap> pix = loadImage(openResourceFile(*dir, file, gfx_suffix_list));
    if (!pix)
        return pix;

    /* Parse modifications */
    while (strRemove(name, "|")) {
        string_t op = strTrim(strFirst(name, "|"));
        if (op.size() == 0) {
            /* Blank */
        } else if (op == "size:screen") {
            /* Resize to screen */
            if (ui_screen) {
                Ptr<GfxPixmap> npix = new GfxPixmap(ui_screen->getSize().x,
                                                    ui_screen->getSize().y,
                                                    *pix);
                doStretchRotateBlit(*pix, *npix,
                                    GfxRect(*pix), GfxRect(*npix),
                                    0, 0,
                                    ui_screen->getSize().x, 0,
                                    0, ui_screen->getSize().y);
                pix = npix;
            }
        } else {
            Ptr<GfxPixmap> ovl = loadImage(openResourceFile(*dir, op, gfx_suffix_list));
            if (ovl)
                ovl->blit(*pix, 0, 0);
        }
    }

    return pix;
}

SfxSampleData*
ResDirectory::loadSound(ResId id)
{
    string_t name = id.toString();
    alias_map_t::iterator i = alias_map.find("s." + name);
    if (i != alias_map.end()) {
        name = i->second.name;
    } else {
        convertName(name);
    }

    Ptr<Stream> st = openResourceFile(*dir, name, sfx_suffix_list);
    if (st) {
        // FIXME: exceptions?
        try {
            return new SfxSampleData(*st);
        }
        catch (...)
        { }
    }
    return 0;
}

Ptr<GfxAnimation>
ResDirectory::loadAnimation(ResId /*id*/)
{
    return 0;
}

Ptr<GfxFont>
ResDirectory::getFont(ResId id)
{
    string_t name = id.toString();
    alias_map_t::iterator i = alias_map.find("f." + name);
    if (i != alias_map.end()) {
        name = i->second.name;
    } else {
        convertName(name);
    }

    Ptr<Stream> s = openResourceFile(*dir, name, fnt_suffix_list);
    if (s)
        return new GfxStdFont(s);
    else
        return 0;
}

Ptr<Stream>
ResDirectory::getStream(ResId id)
{
    string_t name = id.toString();
    alias_map_t::iterator i = alias_map.find(name);
    if (i != alias_map.end()) {
        name = i->second.name;
    } else {
        convertName(name);
    }
    return openResourceFile(*dir, name, null_suffix_list);
}

/** Load alias table from given stream. */
void
ResDirectory::loadAliases(Stream& in)
{
    TextFile tf(in);
    string_t line, l, r;
    while (tf.getLine(line)) {
        /* strip comment */
        string_t::size_type p = line.find_first_of(";#");
        if (p != line.npos)
            line.erase(p);
        line = strTrim(line);
        if (line.empty())
            continue;

        /* must be "normal" line */
        if (strSplit(line, l, r, "=")) {
            alias_map[strTrim(l)].name = strTrim(r);
        } else {
            console.write(LOG_ERROR, format(_("%s:%d: file format error -- line ignored"),
                                            in.getName(), tf.getLineNr()));
        }
    }
}
