/*
 *  TEST: user interface kernel
 *
 *  This displays some silly widgets, to test event propagation,
 *  clipping, redraw etc. While the program runs, keys affect
 *  display:
 *     'n'    open a new window
 *     'q'    (or click onto a UIButton) exit
 *     'u'    SDL_UpdateRect (to check for display bugs)
 *     'r'    redraw screen from scratch (to check for display bugs)
 *     's'    open a new window, and add that event again to SDL's
 *            asynchronous queue
 */

#include <iostream>
#include <SDL.h>
#include "ui/ui.h"
#include "gfx/gfx.h"
#include "ui/window.h"
#include "ui/uiglobal.h"
#include "io/dirs.h"
#include "util/console.h"
#include "gfx/palette.h"
#include "ui/widgets.h"
#include "cpluslib/holder.h"
#include "ui/layout.h"
#include "ui/prefix.h"
#include "ui/msgbox.h"

static SDL_Surface* sfc;

void newWidget();

class Widget : public UIFixedWidget {
    uint32 color;
    uint8 r,g,b;
 public:
    Widget(GfxRect ex, uint32 acol)
        : UIFixedWidget(), color(acol)
        {
            SDL_GetRGB(color, ui_screen->getSystemSurface()->format, &r, &g, &b);
            setExtent(ex);
        }
    void drawContent(GfxCanvas& can);
    bool handleEvent(const UIEvent& e, bool second);
};

void
Widget::drawContent(GfxCanvas& can)
{
    can.drawBar(getExtent(), color, GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
    can.update(getExtent());
}

bool
Widget::handleEvent(const UIEvent& e, bool second)
{
    if (e.type == ev_Key) {
        if(e.key.code == 'n') {
            std::cout << "new\n";
            newWidget();
            return true;
        } else if (e.key.code == 'q') {
            std::cout << "stop\n";
            stop(1);
            return true;
        } else if (e.key.code == 'u') {
            std::cout << "update\n";
            SDL_UpdateRect(sfc, 0, 0, 0, 0);
            return true;
        } else if (e.key.code == 'r') {
            std::cout << "redraw\n";
            std::cout << "--------------------------\n"
                      << ui_root->toString() << "\n"
                      << "--------------------------\n";
            ui_root->drawWidget();
            return true;
        } else if (e.key.code == 's') {
            newWidget();
            SDL_Event e1;
            e1.type = SDL_KEYDOWN;
            e1.key.state = SDL_PRESSED;
            e1.key.keysym.sym = SDLKey('s');
            e1.key.keysym.unicode = 's';
            e1.key.keysym.mod = SDLMod(0);
            SDL_PushEvent(&e1);
            return true;
        } else if (e.key.code == 'i') {
            string_t s = "some string";
            UIInputLine::run(s, 100, "Input Test", "String:", 300, 0);
            return true;
        } else if (e.key.code == 'm') {
            messageBox("it works!", "great!");
            return true;
        }
    } else if (e.type == ev_Idle || e.type == ev_Tick) {
        if (r < 255)
            ++r;
        if (g < 255)
            ++g;
        if (b < 255)
            ++b;
        uint32 c = SDL_MapRGB(ui_screen->getSystemSurface()->format, r, g, b);
        if (c != color) {
            color = c;
            drawWidget(true);
        }
        /* No return here */
    }
    return UIWidget::handleEvent(e, second);
}

ObjectHolder datengrab;

void
newWidget()
{
    static int window_nr = 0;

    UIRoot* g = ui_root.ptr();

    int x = rand() % 500, y = rand() % 300;
    int w = rand() % (640-x) + 100, h = rand() % (480-y) + 50;
    char buf[100];
    GfxRect r(x,y,w,h);

    sprintf(buf, "#%d", ++window_nr);
    UIWindow& win = datengrab.add(new UIWindow(buf, UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIHBoxLayout::instance5);
    win.setExtent(r);
    r.w /= 2;
    r.h /= 5;
    r.x += 20;
    r.y += 30;

    win.add(datengrab.add(new UIPrefixArg()));
    win.add(datengrab.add(new Widget(r, SDL_MapRGB(sfc->format, rand() % 256, rand() % 256, rand() % 256))));
    r.y += r.h + 2;

    UIAbstractButton& b = datengrab.add(new UIButton(1000, "Button", UIKey())).addStop();
    b.setExtent(r);
    win.add(b);
    g->add(win);
}

int
main(int /*argc*/, char** /*argv*/)
{
    if (SDL_Init(SDL_INIT_VIDEO) != 0) {
        printf("error initializing SDL: %s\n", SDL_GetError());
        return 1;
    }
    atexit(SDL_Quit);

    sfc = SDL_SetVideoMode(640, 480, 16, SDL_HWPALETTE);
    if (!sfc) {
        printf("can't set video mode\n");
        return 1;
    }

    initPalette(sfc);
    initUIGlobals(0, "testui");

    newWidget();
    ui_root->run();
}
