/**
  *  \file ui/iconbox.cc
  *  \brief Icon box widget
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "ui/iconbox.h"
#include "gfx/clip.h"
#include "ui/skin.h"

/** Constructor.
    \param size Iconbox size */
UIIconBox::UIIconBox(GfxPoint size)
    : size(size),
      current_item(0),
      hover_item(-1),
      left_x(0),
      target_left_x(0),
      scroll_speed(0),
      pending_scroll(false),
      mouse_pressed(false),
      mouse_blocked(false),
      change_on_click(false)
{ }

void
UIIconBox::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = size;
    info.flags = li_FlexibleH;
}

void
UIIconBox::drawContent(GfxCanvas& can)
{
    int count = getNumberOfItems();
    int x = 0;
    for (int i = 0; i < count; ++i) {
        int wi = getItemWidth(i);
        int itemx = extent.x - left_x + x;
        int itemy = extent.y;
        GfxRect cliprect(itemx, itemy, wi, extent.h);
        cliprect.intersect(extent);
        if (cliprect.exists()) {
            GfxClipRectFilter filter(cliprect, can);
            drawItem(i, filter, GfxRect(itemx, itemy, wi, extent.h),
                     i == current_item ? Selected : i == hover_item ? Hover : Normal);
        }
        x += wi;
    }

    int curx = extent.x - left_x + x;
    if (curx < extent.getRightX()) {
        drawBackground(GfxContext().useCanvas(can).useColorScheme(getSkin()),
                       GfxRect(curx, extent.y, extent.getRightX() - curx, extent.h));
    }

    can.update(extent);
}

void
UIIconBox::onStateChange(int state, bool enable)
{
    if ((state & st_Selected) != 0 && !enable) {
        if (hover_item > 0) {
            hover_item = -1;
            drawWidget();
        }
        if (pending_scroll) {
            if (adjustPosition())
                drawWidget();
            pending_scroll = false;
        }
        mouse_pressed = false;
    }
    UIWidget::onStateChange(state, enable);
}


bool
UIIconBox::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
            /* mouse points into this widget */
            setSelectedWidget();

            /* find item mouse points to */
            int x = 0;
            int new_item = -1;
            int count = getNumberOfItems();
            for (int i = 0; i < count; ++i) {
                int wi = getItemWidth(i);
                if (e.mouse.x >= extent.x - left_x + x && e.mouse.x < extent.x - left_x + x + wi) {
                    new_item = i;
                    break;
                }
                x += wi;
            }

            bool redraw = false;
            if (e.mouse.b != 0) {
                if (change_on_click) {
                    /* just hover */
                    if (new_item != hover_item) {
                        hover_item = new_item;
                        redraw = true;
                    }
                    mouse_pressed = true;
                } else {
                    /* mouse button is pressed, so select that item (but don't scroll,
                       that would immediately select another item) */
                    if (hover_item >= 0) {
                        hover_item = -1;
                        redraw = true;
                    }
                    if (new_item != current_item && new_item >= 0 && !mouse_blocked) {
                        current_item = new_item;
                        mouse_pressed = true;
                        sig_changed.raise(current_item);
                        mouse_pressed = false;
                        redraw = true;
                        pending_scroll = true;
                    }
                }
            } else {
                /* mouse not pressed, just hover */
                mouse_blocked = false;
                if (change_on_click) {
                    if (mouse_pressed) {
                        if (new_item != current_item && new_item >= 0) {
                            current_item = new_item;
                            mouse_pressed = true;
                            sig_changed.raise(current_item);
                            mouse_pressed = false;
                            redraw = true;
                            pending_scroll = true;
                        }
                    }
                    mouse_pressed = false;
                } else {
                    if (new_item != hover_item) {
                        hover_item = new_item;
                        redraw = true;
                    }
                    if (pending_scroll) {
                        /* we've selected a different item during last mouse-click, so
                           we need to scroll */
                        if (adjustPosition())
                            redraw = true;
                        pending_scroll = false;
                    }
                }
            }
            if (redraw) {
                drawWidget();
            }
            return true;
        } else {
            mouse_blocked = false;
            if (hover_item >= 0) {
                hover_item = -1;
                drawWidget();
            }
            return UIWidget::handleEvent(e, second_pass);
        }
    } else if (e.type == ev_Tick) {
        /* Process pending movement */
        if (left_x != target_left_x) {
            /* Adjust speed */
            int distance = std::abs(left_x - target_left_x);
            if (distance > scroll_speed*scroll_speed)
                ++scroll_speed;
            else if (scroll_speed > 1)
                --scroll_speed;

            /* Move. Post a mouse event when reaching the target,
               to update the Hover status */
            if (left_x < target_left_x) {
                left_x += scroll_speed;
                if (left_x >= target_left_x) {
                    left_x = target_left_x;
                    scroll_speed = 0;
                    postMouseEvent();
                }
            } else if (left_x > target_left_x) {
                left_x -= scroll_speed;
                if (left_x <= target_left_x) {
                    left_x = target_left_x;
                    scroll_speed = 0;
                    postMouseEvent();
                }
            }
            drawWidget();
        }
        return UIWidget::handleEvent(e, second_pass);
    } else {
        return UIWidget::handleEvent(e, second_pass);
    }
}

/** Set current item. Also adjusts display so that the item is actually visible.
    \param n Item index [0,getNumberOfItems()). */
void
UIIconBox::setCurrent(int n)
{
    bool redraw_needed = false;
    bool signal_needed = false;

    int total = getNumberOfItems();
    if (total > 0) {
        /* range check */
        if (n >= total)
            n = total-1;
        if (n < 0)
            n = 0;

        /* select the item */
        if (current_item != n) {
            current_item = n;
            signal_needed = true;
            redraw_needed = true;
            adjustPosition();
        }
    } else {
        /* degenerate case */
        if (current_item > 0 || left_x > 0)
            redraw_needed = true;
        current_item = 0;
        left_x = 0;
        target_left_x = 0;
        scroll_speed = 0;
    }

    /* redraw? */
    if (redraw_needed)
        drawWidget();
    if (signal_needed)
        sig_changed.raise(current_item);
}

/** Handle structure change. Reinitializes this iconbox after a complete
    change to the content, and selects a new current item. */
void
UIIconBox::onStructureChange(int n)
{
    /* Set current position */
    setCurrent(n);

    /* Additional adjustments */
    int total = getNumberOfItems();
    if (total > 0) {
        /* compute width before item */
        int before_width = 0;
        for (int i = 0; i < current_item; ++i)
            before_width += getItemWidth(i);
        int item_width = getItemWidth(current_item);
        int additional_width = (current_item >= total-1 ? 0 : 10);

        if (before_width + item_width + additional_width <= extent.w) {
            /* everything fits into the widget */
            target_left_x = 0;
        } else {
            /* does not fit */
            target_left_x = before_width + item_width + additional_width - extent.w;
        }
    }

    /* If mouse is pressed, block it to avoid that it immediately "clicks" again */
    if (mouse_pressed)
        mouse_blocked = true;
}

/** Adjust position so that current item is completely visible.
    \retval true if redraw needed
    \retval false if no change */
bool
UIIconBox::adjustPosition()
{
    /* make sure item is visible */
    int this_item_width = getItemWidth(current_item);
    int this_item_x = 0;
    for (int i = 0; i < current_item; ++i)
        this_item_x += getItemWidth(i);

    int new_left_x = left_x;
    int this_item_right = this_item_x + this_item_width - extent.w;

    if (this_item_x < left_x) {
        /* item is to the left of visible area */
        new_left_x = this_item_x;
    } else if (this_item_x + this_item_width > left_x + extent.w) {
        /* item is to the right of visible area */
        new_left_x = this_item_right;
    }

    /* If new_left_x is at a boundary, and there is an item to the
       left/right of us, move that into visible range */
    if (new_left_x <= this_item_right && current_item != getNumberOfItems()-1) {
        new_left_x = this_item_right + 10;
    }
    if (new_left_x >= this_item_x - 10 && current_item != 0) {
        new_left_x -= 10;
        if (new_left_x < 0)
            new_left_x = 0;
    }

    /* commit change */
    pending_scroll = false;
    if (new_left_x != target_left_x) {
        /* set new X */
        target_left_x = new_left_x;

        /* cancel animation if we're not visible? */
        if (!hasState(st_Exposed)) {
            left_x = target_left_x;
            scroll_speed = 0;
        }

        return true;
    } else {
        return false;
    }
}

