/**
  *  \file ui/widgets/focusgroup.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/focusgroup.h"
#include "ui/layout.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"

/** \class UIFocusGroup
    \brief Group that can have Focus, Visibly

    This is a group which can indicate visibly that it has focus.
    It also ensures that keyboard focus is distributed correctly.
    Otherwise it is exactly like UIGroup.

    \todo Handling of disabled groups is not yet complete.
    - we should implement getCanvasFor() to allow children's
      drawWidget(true) requests. This would require implementation
      of a canvas that performs automatic blending upon update().
      It would make sense to use that in drawEverything() as well.
    - support for drawWidget(false) (see onStateChange()) is a hack */
UIFocusGroup::UIFocusGroup(UILayoutManager& lay, int id, int frame)
    : UILayoutableWidget(lay, id), frame(frame)
{ }

UIFocusGroup::~UIFocusGroup()
{ }

void
UIFocusGroup::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    if (getFocusState() == NoFocus) {
        drawBackground(ctx, GfxRect(extent.x, extent.y, extent.w, 1));
        drawBackground(ctx, GfxRect(extent.x, extent.y+extent.h-1, extent.w, 1));
        drawBackground(ctx, GfxRect(extent.x, extent.y+1, 1, extent.h-2));
        drawBackground(ctx, GfxRect(extent.x+extent.w-1, extent.y+1, 1, extent.h-2));
    } else {
        ctx.setColor(UIColor::tc_Contrast);
        drawRectangle(ctx, extent);
    }
    can.update(extent);
}

void
UIFocusGroup::drawEverything(GfxCanvas& can)
{
    UILayoutableWidget::drawEverything(can);
    if (hasState(st_Disabled)) {
        uint32_t bgcolor = getSkin().getColor(UIColor::tc_Background);
        if (ui_screen_bits < 16) {
            can.drawBar(getExtent(), bgcolor, GFX_TRANSPARENT, GRAY50_FILL, GFX_ALPHA_OPAQUE);
        } else {
            can.drawBar(getExtent(), bgcolor, GFX_TRANSPARENT, SOLID_FILL, 192);
        }
        can.update(getExtent());
    }
}

void
UIFocusGroup::onStateChange(int state, bool enable)
{
    if (state & (st_Focused | st_Disabled))
        drawWidget();

    /* If a child wants to be redrawn, redraw this group completely to make sure
       the "dimming" feature correctly propagates. */
    if (enable && (state & st_ChildRedraw) && hasState(st_Disabled))
        this->status |= st_WantRedraw;

    UILayoutableWidget::onStateChange(state, enable);
}

bool
UIFocusGroup::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Disabled))
        return false;
    if (e.type == ev_Key && !hasState(st_Focused))
        return false;
    if (e.type == ev_Mouse && e.mouse.b
        && !hasState(st_Focused) && isInside(GfxPoint(e.mouse.x, e.mouse.y)))
    {
        setSelectedWidget();
        setFocusedWidget();
        /* FALLTHROUGH */
    }
    return UILayoutableWidget::handleEvent(e, second_pass);
}

void
UIFocusGroup::adjustSize(GfxRect& size, bool to_inner)
{
    if (to_inner)
        size.grow(-frame, -frame);
    else
        size.grow(frame, frame);
}
